"""
Live event engine module.
"""
from queue import Queue, Empty
from threading import Thread
from collections import defaultdict
import logging

from .event import Event

_logger = logging.getLogger(__name__)


class LiveEventEngine(object):
    """
    Event queue + a thread to dispatch events
    """

    def __init__(self):
        """
        Initialize dispatcher thread and handler function list
        """
        # if the dispatcher is active
        self._active = False

        # event queue
        self._queue = Queue()

        # dispatcher thread
        self._thread = Thread(target=self._run)

        # event handlers list, specific event --> handler dict
        self._handlers = defaultdict(list)

    def _run(self):
        """
        run dispatcher
        """
        while self._active == True:
            try:
                event = self._queue.get(block=True, timeout=1)
                # call event handlers
                if event.event_type in self._handlers:
                    for handler in self._handlers[event.event_type]:
                        handler(event)
            except Empty:
                pass
            except Exception as exception:  # pylint: disable=broad-except
                _logger.error("Event %s, Error %s", event.event_type, str(exception))

    def start(self):
        """
        Start the dispatcher thread.
        """
        self._active = True
        self._thread.start()

    def stop(self):
        """
        Stop the dispatcher thread.

        Parameters
        ----------
            timer : bool
                If True, the dispatcher thread will be stopped.
        """
        self._active = False
        self._thread.join()

    def put(self, event: Event) -> None:
        """
        Put event in the queue; call from outside.

        Parameters
        ----------
            event : Event
                Event to put in the queue.
        """
        self._queue.put(event)

    def register_handler(self, type_: str, handler) -> None:
        """
        Register handler/subscriber.

        Parameters
        ----------
            type_ : str
                Event type.

            handler : function
                Handler function.
        """
        handler_list = self._handlers[type_]

        if handler not in handler_list:
            handler_list.append(handler)

    def unregister_handler(self, type_, handler):
        """
        unregister handler/subscriber
        """
        handler_list = self._handlers[type_]

        if handler in handler_list:
            handler_list.remove(handler)

        if not handler_list:
            del self._handlers[type_]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from enum import Enum
from functools import total_ordering


@total_ordering
class OrderStatus(Enum):
    UNKNOWN = 0
    NEWBORN = 1  # in use
    ACKNOWLEDGED = 2  # in use
    PENDING_SUBMIT = 3  # or PRE-SUBMIT
    SUBMITTED = 4  # in use
    PARTIALLY_FILLED = 5
    FILLED = 6  # in use
    PENDING_CANCEL = 7
    CANCELED = 8  # in use
    API_PENDING = 10
    API_CANCELLED = 11
    ERROR = 12

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented



"""
Brokerage base module.
"""
from abc import abstractmethod


class BrokerageBase(object):
    """
    Brokerage base class.
    """

    @abstractmethod
    def place_order(self, order_event):
        """
        Place order.
        """
        raise NotImplementedError("Implement this in your derived class")

    @abstractmethod
    def cancel_order(self, order_id):
        """
        Cancel order.
        """
        raise NotImplementedError("Implement this in your derived class")

    @abstractmethod
    def next_order_id(self):
        """
        Get next order id.
        """
        raise NotImplementedError("Implement this in your derived class")

    # pylint: disable=unused-private-member
    @abstractmethod
    def __calculate_commission(self, full_symbol, fill_price, fill_size):
        """
        Calculate commision. By default it uses IB commission charges.
        """
        raise NotImplementedError("Implement this in your derived class")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .risk_manager_base import RiskManagerBase
import logging

_logger = logging.getLogger(__name__)


class PassThroughRiskManager(RiskManagerBase):
    def order_in_compliance(self, o, strategy_manager=None):
        """
        Pass through the order without constraints
        :param original_order:
        :param env: e.g. strategy_manager that stores order info vs config info
        :return:
        """
        return True


class RiskManager(RiskManagerBase):
    def order_in_compliance(self, o, strategy_manager=None):
        """
        :param original_order:
        :param env: straegy_manager
        :return:
        """
        # 1. check time str hh:mm:ss
        if (
            "order_start_time"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["order_start_time"]
                is not None
            ):
                if (
                    o.create_time
                    < strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["order_start_time"]
                ):
                    _logger.error(
                        f"Order start time breach {o.source}: {o.create_time} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['order_start_time']}"
                    )
                    return False
        if (
            "order_end_time"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["order_end_time"]
                is None
            ):
                if (
                    o.create_time
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["order_end_time"]
                ):
                    _logger.error(
                        f"Order end time breach {o.source}: {o.create_time} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['order_end_time']}"
                    )
                    return False

        # 2. single trade limit; integer
        if (
            "single_trade_limit"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["single_trade_limit"]
                is None
            ):
                if (
                    abs(o.order_size)
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["single_trade_limit"]
                ):
                    _logger.error(
                        f"Order single trade limit breach {o.source}: {o.order_size} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['single_trade_limit']}"
                    )
                    return False

        # total # of trades
        if (
            "total_trade_limit"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["total_trade_limit"]
                is None
            ):
                a = len(
                    strategy_manager._strategy_dict[o.source]._order_manager.order_dict
                ) - len(
                    strategy_manager._strategy_dict[
                        o.source
                    ]._order_manager.canceled_order_set
                )
                if (
                    a
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["total_trade_limit"]
                ):
                    _logger.error(
                        f"Order total trade limit breach {o.source}: {a} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['total_trade_limit']}"
                    )
                    return False
        if "total_trade_limit" in strategy_manager._config.keys():
            if not strategy_manager._config["total_trade_limit"] is None:
                a = len(strategy_manager._order_manager.order_dict) - len(
                    strategy_manager._order_manager.canceled_order_set
                )
                if a > strategy_manager._config["total_trade_limit"]:
                    _logger.error(
                        f"Order global total trade limit breach {o.source}: {a} / {strategy_manager._config['total_trade_limit']}"
                    )
                    return False

        # cancel # limit
        if (
            "total_cancel_limit"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["total_cancel_limit"]
                is None
            ):
                a = len(
                    strategy_manager._strategy_dict[
                        o.source
                    ]._order_manager.canceled_order_set
                )
                if (
                    a
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["total_cancel_limit"]
                ):
                    _logger.error(
                        f"Order total cancel limit breach {o.source}: {a} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['total_cancel_limit']}"
                    )
                    return False
        if "total_cancel_limit" in strategy_manager._config.keys():
            if not strategy_manager._config["total_cancel_limit"] is None:
                a = len(strategy_manager._order_manager.canceled_order_set)
                if a > strategy_manager._config["total_cancel_limit"]:
                    _logger.error(
                        f"Order global total cancel limit breach {o.source}: {a} / {strategy_manager._config['total_cancel_limit']}"
                    )
                    return False

        # active order # limit
        if (
            "total_active_limit"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["total_active_limit"]
                is None
            ):
                a = len(
                    strategy_manager._strategy_dict[
                        o.source
                    ]._order_manager.standing_order_set
                )
                if (
                    a
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["total_active_limit"]
                ):
                    _logger.error(
                        f"Order total active limit breach {o.source}: {a} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['total_active_limit']}"
                    )
                    return False
        if "total_active_limit" in strategy_manager._config.keys():
            if not strategy_manager._config["total_active_limit"] is None:
                a = len(strategy_manager._order_manager.standing_order_set)
                if a > strategy_manager._config["total_active_limit"]:
                    _logger.error(
                        f"Order global total active limit breach {o.source}: {a} / {strategy_manager._config['total_active_limit']}"
                    )
                    return False

        # pnl; note that total loss includes open pnl from existing positions (e.g. bought yesterday, carried overnight)
        if (
            "total_loss_limit"
            in strategy_manager._config["strategy"][
                strategy_manager._strategy_dict[o.source].name
            ].keys()
        ):
            if (
                not strategy_manager._config["strategy"][
                    strategy_manager._strategy_dict[o.source].name
                ]["total_loss_limit"]
                is None
            ):
                a = strategy_manager._strategy_dict[
                    o.source
                ]._position_manager.get_total_pnl() * (-1.0)
                if (
                    a
                    > strategy_manager._config["strategy"][
                        strategy_manager._strategy_dict[o.source].name
                    ]["total_loss_limit"]
                ):
                    _logger.error(
                        f"Order total pnl limit breach {o.source}: {a} / {strategy_manager._config['strategy'][strategy_manager._strategy_dict[o.source].name]['total_loss_limit']}"
                    )
                    return False
        if "total_loss_limit" in strategy_manager._config.keys():
            if not strategy_manager._config["total_loss_limit"] is None:
                a = strategy_manager._position_manager.get_total_pnl() * (-1.0)
                if a > strategy_manager._config["total_loss_limit"]:
                    _logger.error(
                        f"Order global total pnl limit breach {o.source}: {a} / {strategy_manager._config['total_loss_limit']}"
                    )
                    return False

        # TODO check position, or risk reach; maybe not here but periodic check
        return True



from datetime import date, timedelta
import os
import pandas as pd
import pickle
from datetime import datetime


def read_data(
    filepath: str, instrument: str, tz: str = "America/New_York"
) -> pd.DataFrame:
    df = pd.read_csv(filepath, parse_dates=True, sep=",")
    df.set_index("Date", inplace=True)
    df.index = pd.to_datetime(df.index)
    df.index = df.index + pd.DateOffset(hours=16)
    df.index = df.index.tz_localize(tz)  # US/Eastern, UTC
    df = df.loc[df.Instrument == instrument, :]
    return df


def read_ohlcv_csv(filepath, adjust=True, tz="America/New_York"):
    df = pd.read_csv(filepath, header=0, parse_dates=True, sep=",", index_col=0)
    df.index = df.index + pd.DateOffset(hours=16)
    df.index = df.index.tz_localize(tz)  # US/Eastern, UTC
    # df.index = pd.to_datetime(df.index)
    if adjust:
        df["Open"] = df["Adj Close"] / df["Close"] * df["Open"]
        df["High"] = df["Adj Close"] / df["Close"] * df["High"]
        df["Low"] = df["Adj Close"] / df["Close"] * df["Low"]
        df["Volume"] = df["Adj Close"] / df["Close"] * df["Volume"]
        df["Close"] = df["Adj Close"]

    df = df[["Open", "High", "Low", "Close", "Volume"]]
    return df


def read_intraday_bar_pickle(filepath, syms, tz="America/New_York"):
    dict_hist_data = {}
    if os.path.isfile(filepath):
        with open(filepath, "rb") as f:
            dict_hist_data = pickle.load(f)
    dict_ret = {}
    for sym in syms:
        try:
            df = dict_hist_data[sym]
            df.index = df.index.tz_localize(tz)  # # US/Eastern, UTC
            dict_ret[sym] = df
        except:
            pass
    return dict_ret


def read_tick_data_txt(filepath, remove_bo=True, tz="America/New_York"):
    """
    filename = yyyymmdd.txt
    """
    asofdate = filepath.split("/")[-1].split(".")[0]
    data = pd.read_csv(filepath, sep=",", header=None)
    data.columns = [
        "Time",
        "ProcessTime",
        "Ticker",
        "Type",
        "BidSize",
        "Bid",
        "Ask",
        "AskSize",
        "Price",
        "Size",
    ]
    data = data[
        ["Time", "Ticker", "Type", "BidSize", "Bid", "Ask", "AskSize", "Price", "Size"]
    ]
    if remove_bo:
        data = data[data.Type.str.contains("TickType.TRADE")]
    data.Time = data.Time.apply(
        lambda t: datetime.strptime(f"{asofdate} {t}", "%Y%m%d %H:%M:%S.%f")
    )
    data.set_index("Time", inplace=True)
    data.index = data.index.tz_localize(tz)  # # US/Eastern, UTC
    dg = data.groupby("Ticker")
    dict_ret = {}
    for sym, dgf in dg:
        dgf = dgf[~dgf.index.duplicated(keep="last")]
        dict_ret[sym] = dgf
    return dict_ret


def save_one_run_results(output_dir, equity, df_positions, df_trades, batch_tag=None):
    df_positions.to_csv(
        "{}{}{}{}".format(
            output_dir, "/positions_", batch_tag if batch_tag else "", ".csv"
        )
    )
    df_trades.to_csv(
        "{}{}{}{}".format(
            output_dir, "/trades_", batch_tag if batch_tag else "", ".csv"
        )
    )
    equity.to_csv(
        "{}{}{}{}".format(
            output_dir, "/equity_", batch_tag if batch_tag else "", ".csv"
        )
    )



"""
Backtest event engine.
"""
from collections import defaultdict
import logging

from queue import Empty, Queue


_logger = logging.getLogger(__name__)


class BacktestEventEngine(object):
    """
    Event queue + a while loop to dispatch events
    """

    def __init__(self, datafeed):
        """
        Initialize handler function list
        """
        # if the data feed is active
        self._active = True

        # event queue
        self._queue = Queue()

        # pull from backtest data feed
        self._datafeed = datafeed

        # event handlers list, dict: specific event key --> handler value
        self._handlers = defaultdict(list)

    def run(self, n_steps=-1):
        """
        run backtest,
        if n_steps = -1, run to the end; else run n_steps
        """
        _logger.info("Running Backtest...")
        nstep = 0
        while self._active:
            try:
                event = self._queue.get(False)
            except Empty:  # throw good exception
                if (n_steps == -1) or (nstep < n_steps):
                    try:
                        event = self._datafeed.stream_next()
                        self._queue.put(event)
                        nstep += 1
                    except:  # pylint: disable=bare-except
                        # stop if not able to next event
                        self._active = False
                else:
                    break
            else:  # not empty
                try:
                    # call event handlers
                    if event.event_type in self._handlers:
                        for handler in self._handlers[event.event_type]:
                            handler(event)
                except Exception as exception:  # pylint: disable=broad-except
                    logging.error("Error %s", exception.args[0])

    def put(self, event):
        """
        put event in the queue; call from outside
        """
        self._queue.put(event)

    def register_handler(self, type_, handler):
        """
        register handler/subscriber
        """
        handler_list = self._handlers[type_]
        if handler not in handler_list:
            handler_list.append(handler)

    def unregister_handler(self, type_, handler):
        """
        unregister handler/subscriber
        """
        handler_list = self._handlers[type_]
        if handler in handler_list:
            handler_list.remove(handler)
        if not handler_list:
            del self._handlers[type_]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .order_status import OrderStatus
from .order_flag import OrderFlag
from .order_type import OrderType
from ..event.event import *


class OrderEvent(Event):
    """
    Order event
    """

    def __init__(self):
        """
        order and order status
        """
        self.event_type = EventType.ORDER
        self.order_id = -1
        self.order_type = OrderType.MARKET
        self.order_flag = OrderFlag.OPEN
        self.order_status = OrderStatus.UNKNOWN
        self.full_symbol = ""
        self.order_size = 0  # short < 0, long > 0
        self.limit_price = 0.0
        self.stop_price = 0.0
        self.fill_size = 0
        self.fill_price = 0.0
        self.create_time = None
        self.fill_time = None
        self.cancel_time = None
        self.account = ""
        self.source = -1  # sid, -1: unknown, 0: discretionary
        self.timestamp = ""

    def __str__(self):
        return "Time: %s, Source: %s, Type: %s, LMT: %s, STP %s Size %s" % (
            self.timestamp,
            str(self.source),
            str(self.order_type),
            str(self.limit_price),
            str(self.stop_price),
            str(self.order_size),
        )



"""
Live feed module.
"""
from datetime import datetime, timedelta
import logging


import pandas as pd
import quandl

from .data_feed_base import DataFeedBase
from ..data.bar_event import BarEvent


_logger = logging.getLogger(__name__)


# pylint: disable=abstract-method
class LiveDataFeed(DataFeedBase):
    """
    Live DataFeed class
    """

    def __init__(
        self,
        events_queue,
        init_tickers=None,
        start_date=None,
        end_date=None,
        calc_adj_returns=False,
    ):
        """
        Takes the CSV directory, the events queue and a possible
        list of initial ticker symbols then creates an (optional)
        list of ticker subscriptions and associated prices.

        Parameters:
        -----------
            events_queue : Queue
                The Queue object from the backtester.

            init_tickers : list, optional
                The list of ticker symbols to initialize the
                price handler with.

            start_date : datetime, optional
                The start date of the backtest.

            end_date : datetime, optional
                The end date of the backtest.

            calc_adj_returns : bool, optional
                Calculate and store the adjusted closing price
        """
        self.events_queue = events_queue
        self.continue_backtest = True
        self.tickers = {}
        self.tickers_data = {}
        self.start_date = start_date
        self.end_date = end_date

        if init_tickers is not None:
            for ticker in init_tickers:
                self.subscribe_ticker(ticker)

        self.bar_stream = self.__merge_sort_ticker_data()
        self.calc_adj_returns = calc_adj_returns
        if self.calc_adj_returns:
            self.adj_close_returns = []

    def __open_ticker_price_online(self, ticker: str) -> None:
        """
        Opens the CSV online from yahoo finance, then store in a dictionary.

        Parameters:
        -----------
            ticker : str
                The ticker symbol to open the CSV online from yahoo finance.
        """
        if self.end_date is not None:
            end_date = self.end_date
        else:
            end_date = datetime.today()
        if self.start_date is not None:
            start_date = self.start_date
        else:
            start_date = end_date - timedelta(days=365)

        data = quandl.get(
            "wiki/" + ticker,
            start_date=start_date,
            end_date=end_date,
            authtoken="your_token",
        )
        self.tickers_data[ticker] = data
        self.tickers_data[ticker]["Ticker"] = ticker

    def __merge_sort_ticker_data(self) -> None:
        """
        Concatenates all of the separate equities DataFrames
        into a single DataFrame that is time ordered, allowing tick
        data events to be added to the queue in a chronological fashion.

        Note that this is an idealised situation, utilised solely for
        backtesting. In live trading ticks may arrive "out of order".
        """
        dfm = pd.concat(self.tickers_data.values()).sort_index()
        start = None
        end = None
        if self.start_date is not None:
            start = dfm.index.searchsorted(self.start_date)
        if self.end_date is not None:
            end = dfm.index.searchsorted(self.end_date)
        # Determine how to slice
        if start is None and end is None:
            return dfm.iterrows()
        elif start is not None and end is None:
            return dfm.ix[start:].iterrows()
        elif start is None and end is not None:
            return dfm.ix[:end].iterrows()
        else:
            return dfm.ix[start:end].iterrows()

    def subscribe_ticker(self, ticker: str) -> None:
        """
        Subscribes the price handler to a new ticker symbol.

        Parameters
        ----------
            ticker : str
                The ticker symbol to subscribe to.
        """
        if ticker not in self.tickers:
            try:
                self.__open_ticker_price_online(ticker)
                dft = self.tickers_data[ticker]
                row0 = dft.iloc[0]

                close = row0["Close"]
                adj_close = row0["Adj. Close"]

                ticker_prices = {
                    "close": close,
                    "adj_close": adj_close,
                    "timestamp": dft.index[0],
                }
                self.tickers[ticker] = ticker_prices
            except OSError:
                _logger.error(
                    "Could not subscribe ticker %s as no data CSV found for pricing.",
                    ticker,
                )
        else:
            _logger.error(
                "Could not subscribe ticker %s as is already subscribed.", ticker
            )

    @staticmethod
    def __create_event(index, period, ticker, row):
        """
        Obtain all elements of the bar from a row of dataframe
        and return a BarEvent
        """
        open_price = row["Open"]
        high_price = row["High"]
        low_price = row["Low"]
        close_price = row["Close"]
        adj_close_price = row["Adj. Close"]
        volume = int(row["Volume"])
        bev = BarEvent()
        bev.bar_start_time = index
        bev.full_symbol = ticker
        bev.interval = period
        bev.open_price = open_price
        bev.high_price = high_price
        bev.low_price = low_price
        bev.close_price = close_price
        bev.volume = volume
        bev.adj_close_price = adj_close_price
        return bev

    def __store_event(self, event):
        """
        Store price event for closing price and adjusted closing price
        """
        ticker = event.ticker
        # If the calc_adj_returns flag is True, then calculate
        # and store the full list of adjusted closing price
        # percentage returns in a list
        # TODO: Make this faster
        if self.calc_adj_returns:
            prev_adj_close = self.tickers[ticker]["adj_close"]
            cur_adj_close = event.adj_close_price
            self.tickers[ticker]["adj_close_ret"] = cur_adj_close / prev_adj_close - 1.0
            self.adj_close_returns.append(self.tickers[ticker]["adj_close_ret"])
        self.tickers[ticker]["close"] = event.close_price
        self.tickers[ticker]["adj_close"] = event.adj_close_price
        self.tickers[ticker]["timestamp"] = event.time

    def stream_next(self):
        """
        Place the next BarEvent onto the event queue.
        """
        try:
            index, row = next(self.bar_stream)
        except StopIteration:
            self.continue_backtest = False
            return
        # Obtain all elements of the bar from the dataframe
        ticker = row["Ticker"]
        period = 86400  # Seconds in a day
        # Create the tick event for the queue
        bev = self.__create_event(index, period, ticker, row)
        # Store event
        self.__store_event(bev)
        # Send event to queue
        self.events_queue.put(bev)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from trading.v2.data.tick_event import TickEvent
from .brokerage_base import BrokerageBase
from ..data.data_board import DataBoard
from ..event.backtest_event_engine import BacktestEventEngine
from ..event.live_event_engine import LiveEventEngine
from ..order.fill_event import FillEvent
from ..order.order_event import OrderEvent
from ..order.order_type import OrderType
from ..order.order_status import OrderStatus


class BacktestBrokerage(BrokerageBase):
    """
    Market order is immediately filled.
    Limit or stop order is saved to _active_orders for next tick
    """

    def __init__(
        self,
        events_engine: BacktestEventEngine | LiveEventEngine,
        data_board: DataBoard,
    ) -> None:
        """
        Initialize Backtest Brokerage.

        Parameters
        ----------
            events_engine: BacktestEventEngine or LiveEventEngine
                Send fill_event to event engine.

            data_board: DataBoard
                Retrieve latest price from data_board.
        """
        self._events_engine = events_engine
        self._data_board = data_board  # retrieve price against order
        self.orderid = 1
        self.market_data_subscription_reverse_dict = (
            {}
        )  # market data subscription, to be consistent with live
        self._active_orders = {}

    def __calculate_commission(
        self, full_symbol: str, fill_price: float, fill_size: int
    ) -> None:
        """
        Calculate commision. By default it uses IB commission charges.

        Parameters
        ----------
            full_symbol: str
                Contract symbol.

            fill_price: float
                Order fill price.

            fill_size: int
                Order fill size.
        """
        if "STK" in full_symbol:
            commission = max(0.005 * abs(fill_size), 1)  # per share
        elif "FUT" in full_symbol:
            commission = 2.01 * abs(fill_size)  # per contract
        elif "OPT" in full_symbol:
            commission = max(0.7 * abs(fill_size), 1)
        elif "CASH" in full_symbol:
            commission = max(0.000002 * abs(fill_price * fill_size), 2)
        else:
            commission = 0.0001 * abs(
                fill_price * fill_size
            )  # assume 1bps for all other types

        return commission

    def __try_cross_order(self, order_event: OrderEvent, current_price: float) -> None:
        """
        Cross standing order against current price.

        Parameters
        ----------
            order_event: OrderEvent
                Order to be crossed.

            current_price: float
                Current market price.
        """
        if order_event.order_type == OrderType.MARKET:
            order_event.order_status = OrderStatus.FILLED
        # stop limit, if buy, limit price < market price < stop price;
        # if sell, limti price > market price > stop price
        # cross if opposite
        # limit: if buy, limit_price >= current_price; if sell, opposite
        elif (order_event.order_type in [OrderType.LIMIT, OrderType.STOP_LIMIT]) & (
            ((order_event.order_size > 0) & (order_event.limit_price >= current_price))
            | (
                (order_event.order_size < 0)
                & (order_event.limit_price <= current_price)
            )
        ):
            order_event.order_status = OrderStatus.FILLED
        # stop: if buy, stop_price <= current_price; if sell, opposite
        elif (order_event.order_type in [OrderType.STOP, OrderType.STOP_LIMIT]) & (
            ((order_event.order_size > 0) & (order_event.stop_price <= current_price))
            | ((order_event.order_size < 0) & (order_event.stop_price >= current_price))
        ):
            order_event.order_status = OrderStatus.FILLED

    def reset(self) -> None:
        """
        Reset Backtest Brokerage.
        """
        self._active_orders.clear()
        self.orderid = 1

    def on_tick(self, tick_event: TickEvent) -> None:
        """
        Cross standing orders against new tick_event

        Market order can be potentially saved and then filled here against tomorrow's open price

        Parameters
        ----------
            tick_event: TickEvent
                New tick just came in.

        Returns
        -------
            No return; if orders are filled, they are pushed into message queue
        """
        # check standing (stop) orders
        # put trigged into queue
        # and remove from standing order list
        _remaining_active_orders_id = []
        timestamp = tick_event.timestamp
        for oid, order_event in self._active_orders.items():
            # this should be after data board is updated
            # current_price = self._data_board.get_last_price(tick_event.full_symbol)      # last price is not updated yet
            current_price = self._data_board.get_current_price(
                order_event.full_symbol, timestamp
            )
            self.__try_cross_order(order_event, current_price)

            if order_event.order_status == OrderStatus.FILLED:
                fill = FillEvent()
                fill.order_id = order_event.order_id
                fill.fill_id = order_event.order_id
                fill.fill_time = timestamp
                fill.full_symbol = order_event.full_symbol
                fill.fill_size = order_event.order_size
                # TODO: use bid/ask to fill short/long
                fill.fill_price = current_price
                fill.exchange = "BACKTEST"
                fill.commission = self.__calculate_commission(
                    fill.full_symbol, fill.fill_price, fill.fill_size
                )
                self._events_engine.put(fill)
            else:
                # Trailing stop; reset stop price, use limit price as trailing amount
                # if buy, stop price drops when market price drops
                # if sell, stop price increases when market price increases
                if order_event.order_type == OrderType.TRAIING_STOP:
                    if order_event.order_size > 0:
                        order_event.stop_price = min(
                            current_price + order_event.limit_price,
                            order_event.stop_price,
                        )
                    else:
                        order_event.stop_price = max(
                            current_price - order_event.limit_price,
                            order_event.stop_price,
                        )

                _remaining_active_orders_id.append(order_event.order_id)

        self._active_orders = {
            k: v
            for k, v in self._active_orders.items()
            if k in _remaining_active_orders_id
        }

    def place_order(self, order_event: OrderEvent) -> None:
        """
        Place and fill client order; return fill event.

        Market order is immediately filled, no latency or slippage
        the alternative is to save the orders and fill in on_tick function

        Parameters
        ----------
            order_event: OrderEvent
                Client order received.

        Returns
        -------
            No return; fill_event is pushed into message queue.
        """
        # current_price = self._data_board.get_last_price(order_event.full_symbol)      # last price is not updated yet
        timestamp = order_event.create_time
        current_price = self._data_board.get_current_price(
            order_event.full_symbol, timestamp
        )
        self.__try_cross_order(order_event, current_price)

        if order_event.order_status == OrderStatus.FILLED:
            fill = FillEvent()
            fill.order_id = order_event.order_id
            fill.fill_id = order_event.order_id
            # fill.fill_time = self._data_board.get_last_timestamp(order_event.full_symbol)
            fill.fill_time = timestamp
            fill.full_symbol = order_event.full_symbol
            fill.fill_size = order_event.order_size
            # TODO: use bid/ask to fill short/long
            fill.fill_price = current_price
            fill.exchange = "BACKTEST"
            fill.commission = self.__calculate_commission(
                fill.full_symbol, fill.fill_price, fill.fill_size
            )

            order_event.order_status = OrderStatus.FILLED
            self._events_engine.put(order_event)
            self._events_engine.put(fill)
        else:
            order_event.order_status = OrderStatus.ACKNOWLEDGED
            self._active_orders[
                order_event.order_id
            ] = order_event  # save standing orders
            self._events_engine.put(order_event)

    def cancel_order(self, order_id: str) -> None:
        """
        Handle cancel order request from client.

        Parameters
        ----------
            order_id: str
                Order id of the order to be canceled.

        Returns
        -------
            No return; cancel feedback is pushed into message queue.
        """
        self._active_orders = {k: v for k, v in self._active_orders if k != order_id}

    def next_order_id(self) -> str:
        """
        Return next available order id for client to use.

        Returns
        -------
            str: Next available new order id.
        """
        return self.orderid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from enum import Enum

# OrderType.MKT.name == 'MKT'  OderType.MKT.value == 1
class OrderType(Enum):
    UNKNOWN = 0
    MARKET = 1
    LIMIT = 2
    STOP = 3
    STOP_LIMIT = 4
    TRAIING_STOP = 5



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from ..event.event import *
from .position import Position


class PositionEvent(Event):
    """
    position event directly from live broker
    """

    def __init__(self):
        """
        Initialises order
        """
        self.event_type = EventType.POSITION
        self.full_symbol = ""
        self_sec_type = ""
        self.average_cost = 0.0
        self.size = 0
        self.pre_size = 0
        self.freezed_size = 0
        self.realized_pnl = 0.0
        self.unrealized_pnl = 0.0
        self.account = ""
        self.timestamp = ""

    def to_position(self):
        pos = Position(self.full_symbol, self.average_cost, self.size)
        pos.realized_pnl = self.realized_pnl
        pos.unrealized_pnl = self.unrealized_pnl
        pos.account = self.account

        return pos

    def __str__(self):
        return "Ticker: %s, Cost: %s, Size: %s, opl: %s, rpl: %s" % (
            str(self.full_symbol),
            str(self.average_cost),
            str(self.size),
            str(self.unrealized_pnl),
            str(self.realized_pnl),
        )



"""
Backtest data feed.
"""
import pandas as pd

from .data_feed_base import DataFeedBase
from .tick_event import TickEvent


class BacktestDataFeed(DataFeedBase):
    """
    BacktestDataFeed uses PLACEHOLDER to stream_next; actual data comes from data_board.get_hist_price
    This is an easy way to handle multiple sources
    """

    def __init__(self, start_date=None, end_date=None) -> None:
        self._end_date = end_date
        self._start_date = start_date
        self._data_stream = None
        self._data_stream_iter = None

    def set_data_source(self, data: pd.DataFrame):
        """
        Set the data source.

        Parameters
        ----------
            data: pd.DataFrame
                The data source.
        """
        if self._data_stream is None:
            self._data_stream = data.index
        else:
            self._data_stream = self._data_stream.join(
                data.index, how="outer", sort=True
            )

    def subscribe_market_data(self, symbols=None):
        """
        Subscribe to market data.

        Parameters
        ----------
            symbols: list of str
                List of symbols to subscribe to.
        """
        if self._start_date:
            if self._end_date:
                self._data_stream = self._data_stream[
                    (self._data_stream >= self._start_date)
                    & (self._data_stream <= self._end_date)
                ]
            else:
                self._data_stream = self._data_stream[
                    self._data_stream >= self._start_date
                ]

        self._data_stream_iter = self._data_stream.__iter__()

    def unsubscribe_market_data(self, symbols=None):
        """
        Unsubscribe from market data.

        Parameters
        ----------
            symbols: list of str
                List of symbols to unsubscribe from.
        """
        pass

    def stream_next(self):
        """
        Place the next TickEvent into the event queue.
        """
        index = next(self._data_stream_iter)

        t = TickEvent()
        t.full_symbol = "PLACEHOLDER"  # place holders
        t.timestamp = index

        return t



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .fill_event import *
from .order_type import *
from .order_flag import *
from .order_manager import *
from .order_status import *
from .order_type import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .account import *
from .brokerage import *
from .data import *
from .event import *
from .log import *
from .order import *
from .performance import *
from .position import *
from .risk import *
from .strategy import *
from .util import *
from .backtest_engine import BacktestEngine
from .trading_env import TradingEnv
from .portfolio_env import PortfolioEnv
from .version import VERSION as __version__

# https://docs.python-guide.org/writing/logging/
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""
Event module.
"""
from enum import Enum


class EventType(Enum):
    """
    Event type.
    """

    UNDEFINED = -1
    TICK = 0
    BAR = 1
    ORDER = 2
    FILL = 3
    CANCEL = 4
    ORDERSTATUS = 5
    ACCOUNT = 6
    POSITION = 7
    CONTRACT = 8
    HISTORICAL = 9
    TIMER = 10
    LOG = 11


class Event(object):
    """
    Base Event class for event-driven system
    """

    event_type = EventType.UNDEFINED

    @property
    def typename(self):
        """
        Type of event.
        """
        return self.event_type.name


class LogEvent(Event):
    """
    Log event:
    TODO seperate ErrorEvent
    """

    def __init__(self):
        self.event_type = EventType.LOG
        self.timestamp = ""
        self.content = ""



"""
Bar event module.
"""
import pandas as pd
from ..event.event import Event, EventType


class BarEvent(Event):
    # pylint: disable=too-many-instance-attributes
    """
    Bar event, aggregated from TickEvent
    """

    def __init__(self):
        """
        Initialises bar
        """
        self.event_type = EventType.BAR
        self.bar_start_time = pd.Timestamp("1970-01-01", tz="UTC")
        self.interval = 86400  # 1day in secs = 24hrs * 60min * 60sec
        self.full_symbol = ""
        self.open_price = 0.0
        self.high_price = 0.0
        self.low_price = 0.0
        self.close_price = 0.0
        self.adj_close_price = 0.0
        self.volume = 0

    def bar_end_time(self):
        """
        Bar end time.
        """
        # To be consistent with (daily) bar backtest, bar_end_time is set to be bar_start_time
        return self.bar_start_time
        # return self.bar_start_time + pd.Timedelta(seconds=self.interval)

    def __str__(self):
        return (
            f"Time: {self.bar_start_time}, Symbol: {self.full_symbol}, Interval: {self.interval}, "
            f"Open: {self.open_price}, High: {self.high_price}, Low: {self.low_price}, "
            f"Close: {self.close_price}, Adj Close: {self.adj_close_price}, Volume: {self.volume}"
        )



"""
Databoard for trading.
"""
import pandas as pd


class DataBoard(object):
    """
    Data tracker that holds current market data info
    """

    def __init__(self):
        self._hist_data_dict = {}
        self._current_data_dict = {}
        self._current_time = None
        self._placeholder = "PLACEHOLDER"
        self._data_index = None

    def initialize_hist_data(self, data_key: str, data: dict) -> None:
        """
        Initialize historical data.

        Parameters
        ----------
            data_key: str
                Key for the data

            data: dict
                Data dictionary.
        """
        self._hist_data_dict[data_key] = data

    def on_tick(self, tick) -> None:
        """
        On tick event.

        Parameters
        ----------
            tick: TickEvent
                Tick event.
        """
        if tick.full_symbol not in self._current_data_dict:
            self._current_data_dict[tick.full_symbol] = None

        self._current_data_dict[tick.full_symbol] = tick
        self._current_time = tick.timestamp

    def get_last_price(self, symbol: str) -> float:
        """
        Returns last price for a given ticker
        because self._current_time has not been updated by current tick.

        Parameters
        ----------
            symbol: str
                Ticker symbol.
        """
        return self.get_current_price(symbol, self._current_time)

    def get_current_price(self, symbol: str, timestamp: int) -> float:
        """
        Returns the most recent price for a given ticker
        based on current timestamp updated outside of data_board.

        Parameters
        ----------
            symbol: str
                Ticker symbol.

            timestamp: int
                Timestamp.
        """
        if symbol in self._current_data_dict:
            return self._current_data_dict[symbol].price
        elif symbol in self._hist_data_dict:
            return self._hist_data_dict[symbol].loc[timestamp, "Close"]
        elif (
            symbol[:-5] in self._hist_data_dict
        ):  # FUT root symbol e.g. CL, -5 assumes CLZ2020
            return self._hist_data_dict[symbol[:-5]].loc[
                timestamp, symbol
            ]  # column series up to timestamp inclusive
        else:
            return None

    def get_last_timestamp(self, symbol: str) -> int:
        """
        Returns the most recent timestamp for a given ticker.

        Parameters
        ----------
            symbol: str
                Ticker symbol.

        Returns
        -------
            int: Timestamp.
        """
        if symbol in self._current_data_dict:
            return self._current_data_dict[symbol].timestamp
        elif self._placeholder in self._current_data_dict:
            return self._current_data_dict[self._placeholder].timestamp
        else:
            return self._current_time

    def get_current_timestamp(self):
        """
        Retrieve current timestamp.

        Returns
        -------
            int: Current timestamp.
        """
        return self._current_time

    def get_hist_price(self, symbol: str, timestamp: int) -> float:
        """
        Get historical price for a given symbol and timestamp.

        Parameters
        ----------
            symbol: str
                Ticker symbol.

            timestamp: int
                Timestamp.

        Returns
        -------
            float: Price.
        """
        if symbol in self._hist_data_dict:
            return self._hist_data_dict[symbol][:timestamp]  # up to timestamp inclusive
        elif symbol[:-5] in self._hist_data_dict.keys():  # FUT root symbol e.g. CL
            return self._hist_data_dict[symbol[:-5]][symbol][
                :timestamp
            ]  # column series up to timestamp inclusive
        else:
            return None

    def get_hist_sym_time_index(self, symbol: str) -> pd.DatetimeIndex:
        """
        Retrieve historical calendar for a symbol
        this is not look forward.

        Parameters
        ----------
            symbol: str
                Ticker symbol.
        """
        if symbol in self._hist_data_dict:
            return self._hist_data_dict[symbol].index
        elif symbol[:-5] in self._hist_data_dict:  # FUT root symbol e.g. CL
            return self._hist_data_dict[symbol[:-5]].index
        else:
            return None

    def get_hist_time_index(self):
        """
        Retrieve historical calendar this is not look forwward.
        """
        if self._data_index is None:
            for k, v in self._hist_data_dict.items():
                if self._data_index is None:
                    self._data_index = v.index
                else:
                    # pylint: disable=attribute-defined-outside-init
                    self._data_index_data_stream = self._data_index.join(
                        v.index,
                        how="outer",
                        sort=True,
                    )

        return self._data_index



"""
Account manager module.
"""

from .account_event import AccountEvent


class AccountManager(object):
    """
    Account manager class.
    """

    def __init__(self, account_id: str) -> None:
        self._account_id = account_id
        self._account_dict = {}  # account id ==> account
        self.reset()

    def reset(self) -> None:
        """
        Reset the account manager.
        """
        self._account_dict.clear()
        # initialize accounts from server_config.yaml
        account = AccountEvent()
        account.account_id = self._account_id
        account.brokerage = "ib"
        self._account_dict[self._account_id] = account

    def on_account(self, account_event: AccountEvent) -> None:
        """
        Update the account manager with the account event.

        Parameters
        ----------
            account_event: AccountEvent
                Account event.
        """
        if account_event.account_id in self._account_dict:
            self._account_dict[
                account_event.account_id
            ].preday_balance = account_event.preday_balance
            self._account_dict[account_event.account_id].balance = account_event.balance
            self._account_dict[
                account_event.account_id
            ].available = account_event.available
            self._account_dict[
                account_event.account_id
            ].commission = account_event.commission
            self._account_dict[account_event.account_id].margin = account_event.margin
            self._account_dict[
                account_event.account_id
            ].closed_pnl = account_event.closed_pnl
            self._account_dict[
                account_event.account_id
            ].open_pnl = account_event.open_pnl
            self._account_dict[
                account_event.account_id
            ].timestamp = account_event.timestamp
        else:
            self._account_dict[account_event.account_id] = account_event



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import logging

_logger = logging.getLogger(__name__)


class PerformanceManager(object):
    """
    https://www.quantopian.com/docs/api-reference/pyfolio-api-reference
    Record equity, positions, and trades in accordance to pyfolio format
    First date will be the first data start date
    """

    def __init__(self, instrument_meta):
        self._symbols = []
        self.instrument_meta = instrument_meta  # sym ==> meta

        self._equity = None
        self._df_positions = None
        self._df_trades = None

    def add_watch(self, data_key, data):
        if "Close" in data.columns:  # OHLCV
            self._symbols.append(data_key)
        else:  # CLZ20, CLZ21
            self._symbols.extend(data.columns)

    #  or each sid
    def reset(self):
        self._realized_pnl = 0.0
        self._unrealized_pnl = 0.0

        self._equity = pd.Series(dtype=np.float64)  # equity line
        self._equity.name = "total"

        self._df_positions = pd.DataFrame(
            columns=self._symbols + ["cash"], dtype=np.float64
        )
        self._df_trades = pd.DataFrame(
            np.empty(
                0,
                dtype=np.dtype(
                    [("amount", np.int64), ("price", np.float64), ("symbol", np.str)]
                ),
            )
        )
        # self._df_trades.amount = self._df_trades.amount.astype(int)     # pyfolio transactions

    def on_fill(self, fill_event):
        # self._df_trades.loc[fill_event.timestamp] = [fill_event.fill_size, fill_event.fill_price, fill_event.full_symbol]
        self._df_trades = self._df_trades.append(
            pd.DataFrame(
                {
                    "amount": [int(fill_event.fill_size)],
                    "price": [fill_event.fill_price],
                    "symbol": [fill_event.full_symbol],
                },
                index=[fill_event.fill_time],
            )
        )

    def update_performance(self, current_time, position_manager, data_board):
        """
        update previous time/date
        """
        if self._equity.empty:  # no previous day
            self._equity[current_time] = 0.0

        # on a new time/date, calculate the performances for previous time/date
        if current_time != self._equity.index[-1]:
            performance_time = self._equity.index[-1]
        else:
            # When a new data date comes in, it calcuates performances for the previous day
            # This leaves the last date not updated.
            # So we call the update explicitly
            performance_time = current_time

        equity = 0.0
        self._df_positions.loc[performance_time] = [0] * len(self._df_positions.columns)
        for sym, pos in position_manager.positions.items():
            if sym in self.instrument_meta.keys():
                multiplier = self.instrument_meta[sym]["Multiplier"]
            else:
                multiplier = 1

            # data_board (timestamp) hasn't been updated yet
            if pos.size == 0:  # skip empty size
                continue
            equity += pos.size * data_board.get_last_price(sym) * multiplier
            self._df_positions.loc[performance_time, sym] = (
                pos.size * data_board.get_last_price(sym) * multiplier
            )

        self._df_positions.loc[performance_time, "cash"] = position_manager.cash
        self._equity[performance_time] = equity + position_manager.cash
        self._df_positions.loc[performance_time, "total"] = self._equity[
            performance_time
        ]

        if performance_time != current_time:  # not final day
            self._equity[current_time] = 0.0  # add new date
        else:  # final day, re-arrange column order
            self._df_positions = self._df_positions[self._symbols + ["cash"]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import os
import csv
import logging

_logger = logging.getLogger(__name__)

from .trade_recorder_base import AbstractTradeRecorder


class ExampleTradeRecorder(AbstractTradeRecorder):
    """
    A basic compliance module which writes trades to a
    CSV file in the output directory.
    """

    def __init__(self, output_dir):
        """
        Wipe the existing trade log for the day, leaving only
        the headers in an empty CSV.

        It allows for multiple backtests to be run
        in a simple way, but quite likely makes it unsuitable for
        a production environment that requires strict record-keeping.
        """
        self.output_dir = output_dir
        # Remove the previous CSV file
        today = datetime.datetime.utcnow().date()
        self.csv_filename = "tradelog_" + today.strftime("%Y-%m-%d") + ".csv"

        try:
            fname = os.path.expanduser(os.path.join(self.output_dir, self.csv_filename))
            os.remove(fname)
        except (IOError, OSError):
            _logger.error("No tradelog files to clean.")

        # Write new file header
        fieldnames = [
            "timestamp",
            "ticker",
            "action",
            "quantity",
            "exchange",
            "price",
            "commission",
        ]
        fname = os.path.expanduser(os.path.join(self.output_dir, self.csv_filename))
        with open(fname, "a") as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
            writer.writeheader()

    def record_trade(self, fill):
        """
        Append all details about the FillEvent to the CSV trade log.
        """
        fname = os.path.expanduser(os.path.join(self.output_dir, self.csv_filename))
        with open(fname, "a") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(
                [
                    fill.timestamp,
                    fill.ticker,
                    fill.action,
                    fill.quantity,
                    fill.exchange,
                    round(fill.price, 4),
                    round(fill.commission, 4),
                ]
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..event.event import *


class ContractEvent(Event):
    """
    also serve as contract
    """

    def __init__(self):
        self.event_type = EventType.CONTRACT
        self.full_symbol = ""
        self.local_name = ""
        self.mininum_tick = ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
from datetime import datetime
from ..order.order_event import OrderEvent
from ..order.order_type import OrderType
from ..order import OrderManager
from ..position import PositionManager
import logging

_logger = logging.getLogger(__name__)


class StrategyBase(metaclass=ABCMeta):
    """
    Base strategy class
    """

    def __init__(self):
        """
        initialize trategy
        :param symbols:
        :param events_engine:backtest_event_engine or live_event engine that provides queue_.put()
        """
        self.id = -1  # id
        self.name = ""  # name
        self.symbols = []  # symbols interested
        self.strategy_manager = None  # to place order through strategy_manager
        self._data_board = None  # to get current data
        self._position_manager = PositionManager(
            self.name
        )  # track local positions and cash
        self._order_manager = OrderManager(
            self.name
        )  # manage local (standing) orders and fills

        self.active = False
        self.initialized = False

    def set_capital(self, capital):
        self._position_manager.set_capital(capital)

    def set_symbols(self, symbols):
        self.symbols = symbols

    def set_name(self, name):
        self.name = name
        self._position_manager.name = name
        self._order_manager.name = name

    def set_params(self, params_dict=None):
        if params_dict is not None:
            for key, value in params_dict.items():
                try:
                    self.__setattr__(key, value)
                except:
                    pass

    def on_init(self, strategy_manager, data_board, instrument_meta):
        self.strategy_manager = strategy_manager
        self._data_board = data_board
        self._position_manager.set_instrument_meta(instrument_meta)
        self._position_manager.reset()
        self.initialized = True

    def on_start(self):
        self.active = True

    def on_stop(self):
        self.active = False

    def on_tick(self, tick_event):
        """
        Respond to tick
        """
        # for live trading, turn off p&l tick by not calling super.on_tick()
        # for backtest, call super().on_tick() if need to track positions or npv or cash
        self._position_manager.mark_to_market(
            tick_event.timestamp,
            tick_event.full_symbol,
            tick_event.price,
            self._data_board,
        )

    def on_order_status(self, order_event):
        """
        on order acknowledged
        :return:
        """
        # raise NotImplementedError("Should implement on_order_status()")
        self._order_manager.on_order_status(order_event)

    def on_fill(self, fill_event):
        """
        on order filled
        derived class call super().on_fill first
        """
        self._position_manager.on_fill(fill_event)
        self._order_manager.on_fill(fill_event)

    def place_order(self, o):
        """
        expect user to set up order type, order size and order price
        """
        o.source = self.id  # identify source
        if o.create_time is None:
            o.create_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        if self.active:
            self.strategy_manager.place_order(o)

    def adjust_position(self, sym, size_from, size_to, timestamp=None):
        """
        use market order to adjust position
        :param sym:
        :param size_from:
        :param size_to:
        :param timestamp: used by backtest broker to get price on timestamp
        :return:
        """
        if size_from == size_to:
            return
        o = OrderEvent()
        o.full_symbol = sym
        o.order_type = OrderType.MARKET
        o.order_size = size_to - size_from
        if timestamp is not None:
            o.create_time = timestamp

        self.place_order(o)

    def cancel_order(self, oid):
        if oid in self._order_manager.standing_order_set:
            self._order_manager.on_cancel(oid)
            self.strategy_manager.cancel_order(oid)
        else:
            _logger.error(f"Not a standing order to be cancelled, sid {id}, oid {oid}")

    def cancel_all(self):
        """
        cancel all standing orders from this strategy id
        :return:
        """
        for oid in self._order_manager.standing_order_set:
            self._order_manager.on_cancel(oid)
            self.strategy_manager.cancel_order(oid)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pandas import Timestamp
from ..event.event import *
from ..position.position import Position


class FillEvent(Event):
    """
    Fill event, with filled quantity/size and price
    """

    def __init__(self):
        """
        Initialises fill
        """
        self.event_type = EventType.FILL
        self.order_id = -1
        self.fill_id = -1
        self.full_symbol = ""
        self.fill_time = ""
        self.fill_price = 0.0
        self.fill_size = 0  # size < 0 means short order is filled
        self.exchange = ""
        self.commission = 0.0
        self.account = ""
        self.source = -1
        self.api = ""

    def to_position(self, multiplier=1):
        """
        if there is no existing position for this symbol, this fill will create a new position
        (otherwise it will be adjusted to exisitng position)
        """
        if self.fill_size > 0:
            average_price_including_commission = (
                self.fill_price + self.commission / multiplier
            )
        else:
            average_price_including_commission = (
                self.fill_price - self.commission / multiplier
            )

        new_position = Position(
            self.full_symbol, average_price_including_commission, self.fill_size
        )
        return new_position

    def __str__(self):
        return (
            "Time: %s, Source: %s, Oid: %s, Ticker: %s, Price: %s, Size %s Comm %s"
            % (
                self.fill_time,
                str(self.source),
                str(self.order_id),
                self.full_symbol,
                str(self.fill_price),
                str(self.fill_size),
                str(self.commission),
            )
        )



# pylint: disable=too-many-lines
"""
Interactive Brokers module.
"""
from datetime import datetime
from copy import copy
import logging
from threading import Thread

from ibapi.wrapper import EWrapper
from ibapi.client import EClient
from ibapi import utils

from ibapi.common import (
    BarData,
    FaDataType,
    HistogramDataList,
    ListOfContractDescription,
    ListOfDepthExchanges,
    ListOfFamilyCode,
    ListOfHistoricalTick,
    ListOfHistoricalTickBidAsk,
    ListOfHistoricalTickLast,
    ListOfNewsProviders,
    ListOfPriceIncrements,
    OrderId,
    SetOfFloat,
    SetOfString,
    SmartComponentMap,
    TickAttrib,
    TickAttribBidAsk,
    TickAttribLast,
    TickerId,
)
from ibapi.contract import ComboLeg, Contract, ContractDetails
from ibapi.order import Order

from ibapi.order_state import OrderState
from ibapi.execution import Execution
from ibapi.commission_report import CommissionReport
from ibapi.ticktype import TickTypeEnum

from .brokerage_base import BrokerageBase
from ..event.event import LogEvent
from ..account import AccountEvent
from ..data import TickEvent, TickType, BarEvent
from ..order.order_type import OrderType
from ..order.fill_event import FillEvent
from ..order.order_event import OrderEvent
from ..order.order_status import OrderStatus
from ..position.position_event import PositionEvent


_logger = logging.getLogger(__name__)


class InteractiveBrokers(BrokerageBase):
    """
    Interactive Brokers class.
    """

    def __init__(self, msg_event_engine, tick_event_engine, account) -> None:
        """
        Initialize InteractiveBrokers brokerage.

        Currently, the client is strongly coupled to broker without an incoming queue,
        e.g. client calls broker.place_order to place order directly.

        Parameters
        ----------
            msg_event_engine:
                Used to broadcast messages the broker generates back to client.

            tick_event_engine:
                Used to broadcast market data back to client.

            account:
                The IB account.
        """
        self.event_engine = msg_event_engine  # save events to event queue
        self.tick_event_engine = tick_event_engine
        self.api = IBApi(self)
        self.account = account
        self.contract_detail_request_contract_dict = {}  # reqid ==> contract
        self.contract_detail_request_symbol_dict = {}  # reqid ==> symbol
        self.sym_contract_dict = {}  # sym ==> contract
        self.contract_symbol_dict = {}  # conId ==> symbol
        self.market_data_subscription_dict = {}  # reqId ==> sym
        self.market_data_subscription_reverse_dict = {}  # sym ==> reqId
        self.market_data_tick_dict = (
            {}
        )  # reqid ==> tick_event; to combine tickprice and ticksize
        self.market_depth_subscription_dict = {}
        self.market_depth_subscription_reverse_dict = {}
        self.market_depth_tick_dict = {}  # to combine tickprice and ticksize
        self.hist_data_request_dict = {}
        self.order_dict = {}  # order id ==> order_event
        self.account_summary_reqid = -1
        self.account_summary = AccountEvent()
        self.account_summary.brokerage = "IB"
        self.clientid = 0
        self.reqid = 0  # next/available reqid
        self.orderid = 0  # next/available orderid

    def connect(
        self, host: str = "127.0.0.1", port: int = 7497, client_id: int = 0
    ) -> None:
        """
        Connect to IB. Request open orders under clientid upon successful connection.

        Parameters
        ----------
            host: str
                Host address.

            port: int
                Socket port.

            client_id: int
                Client id.
        """
        self.clientid = client_id
        if self.api.connected:
            return

        self.api.connect(host, port, clientId=client_id)
        self.api.thread.start()
        self.req_current_time()

        if client_id == 0:
            # associate TWS with the client
            self.api.reqAutoOpenOrders(True)

    def disconnect(self) -> None:
        """
        Disconnect from IB
        """
        if not self.api.isConnected():
            return

        self.api.connected = False
        # self.api.conn.disconnect()
        self.api.conn.socket = None
        self.api.disconnect()
        _logger.info("connected %s", self.api.isConnected())

    # pylint: disable=unused-private-member
    def __calculate_commission(self, full_symbol, fill_price, fill_size):
        pass

    def next_order_id(self) -> int:
        """
        Return next available order id

        Returns
        -------
            int: Next order id available for next orders.
        """
        return self.orderid

    def place_order(self, order_event: OrderEvent) -> None:
        """
        Place order to IB

        Parameters
        ----------
            order_event: OrderEvent
                Client order to be placed.

        Returns
        -------
            No return. An order event is pushed to message queue with order status Acknowledged.
        """
        if not self.api.connected:
            return

        ib_contract = InteractiveBrokers.symbol_to_contract(order_event.full_symbol)
        if not ib_contract:
            _logger.error(
                "Failed to find contract to place order %s", order_event.full_symbol
            )
            return

        ib_order = InteractiveBrokers.order_to_ib_order(order_event)
        if not ib_order:
            _logger.error("Failed to create order to place %s", order_event.full_symbol)
            return

        # The EtradeOnly IBApi.Order attribute is no longer supported.
        # Error received with TWS versions 983+
        ib_order.eTradeOnly = False
        # The firmQuoteOnly IBApi.Order attribute is no longer supported.
        # Error received with TWS versions 983+
        ib_order.firmQuoteOnly = False
        if order_event.order_id < 0:
            order_event.order_id = self.orderid
            self.orderid += 1
        order_event.account = self.account
        order_event.timestamp = datetime.now().strftime("%H:%M:%S.%f")
        order_event.order_status = OrderStatus.ACKNOWLEDGED  # acknowledged
        self.order_dict[order_event.order_id] = order_event
        _logger.info(
            "Order acknowledged %s, %s", order_event.order_id, order_event.full_symbol
        )
        self.event_engine.put(copy(order_event))
        self.api.placeOrder(order_event.order_id, ib_contract, ib_order)

    def cancel_order(self, order_id: int) -> None:
        """
        Cancel client order.

        Parameters
        ----------
            order_id: int
                order id of the order to be canceled.

        Returns
        -------
            No return. If order is successfully canceled, IB will return an orderstatus message.
        """
        if not self.api.connected:
            return

        if not order_id in self.order_dict:
            _logger.error("Order to cancel not found. order id %s", order_id)
            return

        self.order_dict[order_id].cancel_time = datetime.now().strftime("%H:%M:%S.%f")

        self.api.cancelOrder(order_id)

    def cancel_all_orders(self) -> None:
        """
        Cancel all standing orders, for example, before one wants
        to shut down completely for some reasons.
        """
        self.api.reqGlobalCancel()

    def subscribe_market_data(self, sym: str) -> None:
        """
        Subscribe market L1 data. Market data for this symbol will then be streamed to client.

        Parameters
        ----------
            sym: str
                The symbol to be subscribed.
        """
        if not self.api.connected:
            return

        # it's not going to re-subscribe, because we only call subscribe_market_datas
        # if sym in self.market_data_subscription_reverse_dict.keys():
        #     return

        contract = InteractiveBrokers.symbol_to_contract(sym)
        if not contract:
            _logger.error("Failed to find contract to subscribe market data: %s", sym)
            return

        self.api.reqContractDetails(self.reqid, contract)
        _logger.info("Requesting market data %s %s", self.reqid, sym)
        self.contract_detail_request_contract_dict[self.reqid] = contract
        self.contract_detail_request_symbol_dict[self.reqid] = sym
        self.reqid += 1
        self.api.reqMktData(self.reqid, contract, "", False, False, [])
        tick_event = TickEvent()
        tick_event.full_symbol = sym
        self.market_data_subscription_dict[self.reqid] = sym
        self.market_data_subscription_reverse_dict[sym] = self.reqid
        self.market_data_tick_dict[self.reqid] = tick_event
        self.reqid += 1

    def subscribe_market_datas(self) -> None:
        """
        Subscribe market L1 data for all symbols used in strategies.
        Market data for this symbol will then be streamed to client.
        """
        syms = list(self.market_data_subscription_reverse_dict.keys())
        for sym in syms:
            self.subscribe_market_data(sym)

    def unsubscribe_market_data(self, sym: str) -> None:
        """
        Unsubscribe market L1 data. Market data for this symbol will stop streaming to client.

        Parameters
        ----------
            sym: str
                The symbol to be subscribed.
        """
        if not self.api.connected:
            return

        if not sym in self.market_data_subscription_reverse_dict:
            return

        self.api.cancelMktData(self.market_data_subscription_reverse_dict[sym])

    def subscribe_market_depth(self, sym: str) -> None:
        """
        Subscribe market L2 data. Market data for this symbol will then be streamed to client.

        Parameters
        ----------
            sym: str
                The symbol to be subscribed.
        """
        if not self.api.connected:
            return

        if sym in self.market_depth_subscription_reverse_dict:
            return

        contract = InteractiveBrokers.symbol_to_contract(sym)
        if not contract:
            _logger.error("Failed to find contract to subscribe market depth: %s", sym)
            return

        self.api.reqMktDepth(self.reqid, contract, 5, True, [])
        self.reqid += 1
        self.market_depth_subscription_dict[self.reqid] = sym
        self.market_depth_subscription_reverse_dict[sym] = self.reqid

    def unsubscribe_market_depth(self, sym) -> None:
        """
        Unsubscribe market L2 data. Market data for this symbol will stop streaming to client.

        Parameters
        ----------
            sym: str
                The symbol to be subscribed.
        """
        if not self.api.connected:
            return

        if not sym in self.market_depth_subscription_reverse_dict:
            return

        self.api.cancelMktDepth(self.market_depth_subscription_reverse_dict[sym], True)

    def subscribe_account_summary(self) -> None:
        """
        Request account summary from broker
        """
        if not self.api.connected:
            return

        if self.account_summary_reqid > 0:  # subscribed
            return

        self.account_summary_reqid = self.reqid
        self.api.reqAccountSummary(self.account_summary_reqid, "All", "$LEDGER")
        self.reqid += 1

    def unsubscribe_account_summary(self) -> None:
        """
        Stop receiving account summary from broker
        """
        if not self.api.connected:
            return

        if self.account_summary_reqid == -1:
            return

        self.api.cancelAccountSummary(self.account_summary_reqid)
        self.account_summary_reqid = -1

    def subscribe_positions(self) -> None:
        """
        Request existing positions from broker
        """
        self.api.reqPositions()

    def unsubscribe_positions(self) -> None:
        """
        Stop receiving existing position message from broker.
        """
        self.api.cancelPositions()

    def request_historical_data(self, symbol: str, end: datetime = None) -> None:
        """
        Request 1800 S (30 mins) historical bar data from Interactive Brokers.

        Parameters
        ----------
            symbol: str
                The contract whose historical data is requested.
            end: datetime
                The end time of the historical data.

        Returns
        -------
            No returns; data is broadcasted through message queue.
        """
        ib_contract = InteractiveBrokers.symbol_to_contract(symbol)

        if end:
            end_str = end.strftime("%Y%m%d %H:%M:%S")
        else:
            end_str = ""

        self.hist_data_request_dict[self.reqid] = symbol
        self.api.reqHistoricalData(
            self.reqid,
            ib_contract,
            end_str,
            "1800 S",
            "1 secs",
            "TRADES",
            1,
            1,
            False,
            [],
        )  # first 1 is useRTH
        self.reqid += 1

    def cancel_historical_data(self, reqid: int) -> None:
        """
        Cancel historical data request. Usually not necessary.

        Parameters
        ----------
            reqid: int
                The historical data request id.
        """
        self.api.cancelHistoricalData(reqid)

    def request_historical_ticks(
        self, symbol: str, start_time: str, reqtype="TICKS"
    ) -> None:
        """
        Request historical time and sales data from Interactive Brokers.
        See here https://interactivebrokers.github.io/tws-api/historical_time_and_sales.html

        Parameters
        ----------
            symbol: str
                The contract whose historical data is requested.

            start_time: str
                i.e. "20170701 12:01:00". Uses TWS timezone specified at login.

            reqtype: str
                TRADES, BID_ASK, or MIDPOINT

        Returns
        -------
            No returns; data is broadcasted through message queue.
        """
        ib_contract = InteractiveBrokers.symbol_to_contract(symbol)
        use_rth = 1
        self.hist_data_request_dict[self.reqid] = symbol
        self.api.reqHistoricalTicks(
            self.reqid, ib_contract, start_time, "", 1000, reqtype, use_rth, True, []
        )
        self.reqid += 1

    def req_current_time(self) -> None:
        """
        Request server time on broker side
        """
        self.api.reqCurrentTime()

    def set_server_log_level(self, level: int = 1) -> None:
        """
        Set server side log level or the log messages received from server.

        Parameters
        ----------
            level: int
                Log level. 1: error, 2: warning, 3: information, 4: detail.
        """
        self.api.setServerLogLevel(level)

    def heartbeat(self) -> None:
        """
        Request server time as heartbeat
        """
        if self.api.isConnected():
            _logger.info("reqPositions")
            # self.api.reqPositions()
            self.req_current_time()  # EWrapper::currentTime

    def log(self, msg: str) -> None:
        """
        Broadcast server log message through message queue

        Parameters
        ----------
            msg: str
                Message to be broadcast.

        Returns
        -------
            No return; log meesage is placed into message queue.
        """
        timestamp = datetime.now().strftime("%H:%M:%S.%f")
        log_event = LogEvent()
        log_event.timestamp = timestamp
        log_event.content = msg
        self.event_engine.put(log_event)

    @staticmethod
    def symbol_to_contract(symbol: str) -> None:
        """
        Convert fulll symbol string to IB contract

        TODO
        CL.HO BAG 174230608 1 NYMEX 257430162 1 NYMEX NYMEX     # Inter-comdty
        ES.NQ BAG 371749798 1 GLOBEX 371749745 1 GLOBEX GLOBEX     # Inter-comdty
        CL.HO BAG 257430162 1 NYMEX 174230608 1 NYMEX NYMEX

        Parameters
        ----------
            symbol: str
                Full symbol, e.g. AMZN STK SMART.

        Returns
        -------
            IB contract.
        """
        symbol_fields = symbol.split(" ")
        ib_contract = Contract()

        if symbol_fields[1] == "STK":
            ib_contract.localSymbol = symbol_fields[0]
            ib_contract.secType = symbol_fields[1]
            ib_contract.currency = "USD"
            ib_contract.exchange = symbol_fields[2]
        elif symbol_fields[1] == "CASH":
            ib_contract.symbol = symbol_fields[0][0:3]  # EUR
            ib_contract.secType = symbol_fields[1]  # CASH
            ib_contract.currency = symbol_fields[0][3:]  # GBP
            ib_contract.exchange = symbol_fields[2]  # IDEALPRO
        elif symbol_fields[1] == "FUT":
            ib_contract.localSymbol = symbol_fields[0].replace(
                "_", " "
            )  # ESM9, in case YM___SEP_20
            ib_contract.secType = symbol_fields[1]  # FUT
            ib_contract.exchange = symbol_fields[2]  # GLOBEX
            ib_contract.currency = "USD"
        elif symbol_fields[1] == "OPT":  # AAPL OPT 20201016 128.75 C SMART
            ib_contract.symbol = symbol_fields[0]  # AAPL
            ib_contract.secType = symbol_fields[1]  # OPT
            ib_contract.lastTradeDateOrContractMonth = symbol_fields[2]  # 20201016
            ib_contract.strike = (
                float(symbol_fields[3])
                if "." in symbol_fields[3]
                else int(symbol_fields[3])
            )  # 128.75
            ib_contract.right = symbol_fields[4]  # C
            ib_contract.exchange = symbol_fields[5]  # SMART
            ib_contract.currency = "USD"
            ib_contract.multiplier = "100"
        elif symbol_fields[1] == "FOP":  # ES FOP 20200911 3450 C 50 GLOBEX
            ib_contract.symbol = symbol_fields[0]  # ES
            ib_contract.secType = symbol_fields[1]  # FOP
            ib_contract.lastTradeDateOrContractMonth = symbol_fields[2]  # 20200911
            ib_contract.strike = (
                float(symbol_fields[3])
                if "." in symbol_fields[3]
                else int(symbol_fields[3])
            )  # 128.75
            ib_contract.right = symbol_fields[4]  # C
            ib_contract.multiplier = symbol_fields[5]  # 50
            ib_contract.exchange = symbol_fields[6]  # GLOBEX
            ib_contract.currency = "USD"
        elif symbol_fields[1] == "CMDTY":  # XAUUSD CMDTY SMART
            ib_contract.symbol = symbol_fields[0]  # XAUUSD
            ib_contract.secType = symbol_fields[1]  # COMDTY
            ib_contract.currency = "USD"
            ib_contract.exchange = symbol_fields[2]  # SMART
        elif symbol_fields[1] == "BAG":
            ib_contract.symbol = symbol_fields[0]  # CL.BZ
            ib_contract.secType = symbol_fields[1]  # BAG

            leg1 = ComboLeg()
            leg1.conId = int(symbol_fields[2])  # 174230608
            leg1.ratio = int(symbol_fields[3])  # 1
            leg1.action = "BUY"
            leg1.exchange = symbol_fields[4]  # NYMEX

            leg2 = ComboLeg()
            leg2.conId = int(symbol_fields[5])  # 162929662
            leg2.ratio = int(symbol_fields[6])  # 1
            leg2.action = "SELL"
            leg2.exchange = symbol_fields[7]  # NYMEX

            ib_contract.comboLegs = []
            ib_contract.comboLegs.append(leg1)
            ib_contract.comboLegs.append(leg2)

            ib_contract.exchange = symbol_fields[8]  # NYMEX
            ib_contract.currency = "USD"
        else:
            _logger.error("invalid contract %s", symbol)

        return ib_contract

    @staticmethod
    def contract_to_symbol(ib_contract: Contract) -> str:
        """
        Convert IB contract to full symbol

        Parameters
        ----------
            ib_contract: Contract
                IB contract.

        Returns
        -------
            str: Full symbol.
        """
        full_symbol = ""
        if ib_contract.secType == "STK":
            full_symbol = " ".join(
                [ib_contract.localSymbol, "STK", "SMART"]
            )  # or ib_contract.primaryExchange?
        elif ib_contract.secType == "CASH":
            full_symbol = " ".join(
                [
                    ib_contract.symbol + ib_contract.currency,
                    "CASH",
                    ib_contract.exchange,
                ]
            )
        elif ib_contract.secType == "FUT":
            full_symbol = " ".join(
                [
                    ib_contract.localSymbol.replace(" ", "_"),
                    "FUT",
                    ib_contract.primaryExchange
                    if ib_contract.primaryExchange != ""
                    else ib_contract.exchange,
                ]
            )
        elif ib_contract.secType == "OPT":
            full_symbol = " ".join(
                [
                    ib_contract.symbol,
                    "OPT",
                    ib_contract.lastTradeDateOrContractMonth,
                    str(ib_contract.strike),
                    ib_contract.right,
                    "SMART",
                ]
            )
        elif ib_contract.secType == "FOP":
            full_symbol = " ".join(
                [
                    ib_contract.symbol,
                    "FOP",
                    ib_contract.lastTradeDateOrContractMonth,
                    str(ib_contract.strike),
                    ib_contract.right,
                    ib_contract.multiplier,
                    ib_contract.exchange,
                ]
            )
        elif ib_contract.secType == "COMDTY":
            full_symbol = " ".join([ib_contract.symbol, "COMDTY", "SMART"])
        elif ib_contract.secType == "BAG":
            full_symbol = " ".join([ib_contract.symbol, "COMDTY", "SMART"])

        return full_symbol

    @staticmethod
    def order_to_ib_order(order_event: OrderEvent) -> Order:
        """
        Convert order event to IB order

        Parameters
        ----------
            order_event: OrderEvent
                Internal representation of order.

        Returns
        -------
            Order: IB representation of order.
        """
        ib_order = Order()
        ib_order.action = "BUY" if order_event.order_size > 0 else "SELL"
        ib_order.totalQuantity = abs(order_event.order_size)
        if order_event.order_type == OrderType.MARKET:
            ib_order.orderType = "MKT"
        elif order_event.order_type == OrderType.LIMIT:
            ib_order.orderType = "LMT"
            ib_order.lmtPrice = order_event.limit_price
        elif order_event.order_type == OrderType.STOP:
            ib_order.orderType = "STP"
            ib_order.auxPrice = order_event.stop_price
        elif order_event.order_type == OrderType.STOP_LIMIT:
            ib_order.orderType = "STP LMT"
            ib_order.lmtPrice = order_event.limit_price
            ib_order.auxPrice = order_event.stop_price
        else:
            return None

        return ib_order

    @staticmethod
    def ib_order_to_order(ib_order: Order) -> OrderEvent:
        """
        Convert IB order to order event

        Parameters
        ----------
            ib_order: Order
                IB representation of order.

        Returns
        -------
            OrderEvent: Internal representation of order.
        """
        order_event = OrderEvent()
        # order_event.order_id = orderId
        # order_event.order_status = orderState.status
        direction = 1 if ib_order.action == "BUY" else -1
        order_event.order_size = ib_order.totalQuantity * direction
        if ib_order.orderType == "MKT":
            order_event.order_type = OrderType.MARKET
        elif ib_order.orderType == "LMT":
            order_event.order_type = OrderType.LIMIT
            order_event.limit_price = ib_order.lmtPrice
        elif ib_order.orderType == "STP":
            order_event.order_type = OrderType.STOP
            order_event.stop_price = ib_order.auxPrice
        elif ib_order.orderType == "STP LMT":
            order_event.order_type = OrderType.STOP_LIMIT
            order_event.limit_price = ib_order.lmtPrice
            order_event.stop_price = ib_order.auxPrice
        else:
            order_event.order_type = OrderType.UNKNOWN
            order_event.limit_price = ib_order.lmtPrice
            order_event.stop_price = ib_order.auxPrice

        return order_event


class IBApi(EWrapper, EClient):
    """
    Interactive Brokers API wrapper.
    """

    def __init__(self, broker):
        EWrapper.__init__(self)
        EClient.__init__(self, wrapper=self)

        self.broker = broker
        self.thread = Thread(target=self.run)

        self.nKeybInt = 0  # pylint: disable=invalid-name
        self.connected = False
        self.globalCancelOnly = False  # pylint: disable=invalid-name
        self.simplePlaceOid = None  # pylint: disable=invalid-name

    def keyboardInterrupt(self) -> None:
        """
        Keyboard interrupt handler.
        """
        self.nKeybInt += 1
        if self.nKeybInt == 1:
            self.stop()
        else:
            _logger.info("Finishing test")

    def stop(self) -> None:
        """
        Stop the IB API.
        """
        _logger.info("Executing cancels")
        _logger.info("Executing cancels ... finished")

    def connectAck(self) -> None:
        if self.asynchronous:
            self.startApi()
        self.connected = True
        _logger.info("IB connected")

    def nextValidId(self, orderId: int) -> None:
        super().nextValidId(orderId)
        msg = f"nextValidOrderId: {orderId}"
        _logger.info(msg)
        self.broker.log(msg)
        self.broker.orderid = orderId

        # we can start now
        self.broker.subscribe_market_datas()

    def error(self, reqId: TickerId, errorCode: int, errorString: str) -> None:
        super().error(reqId, errorCode, errorString)
        msg = f"Error. id: {reqId}, Code: {errorCode}, Msg: {errorString}"
        _logger.error(msg)
        self.broker.log(msg)

    def winError(self, text: str, lastError: int):
        """
        Win error.

        Parameters
        ----------
            text: str
                Error text.

            lastError: int
                Last error.
        """
        super().winError(text, lastError)
        msg = f"Error Id: {lastError}, Msg: {text}"
        _logger.error(msg)
        self.broker.log(msg)

    def openOrder(
        self, orderId: OrderId, contract: Contract, order: Order, orderState: OrderState
    ) -> None:
        super().openOrder(orderId, contract, order, orderState)
        msg = (
            f"OpenOrder. PermId: {order.permId}, ClientId:  {order.clientId}, OrderId: {orderId}, "
            f"Account: {order.account}, Symbol: {contract.symbol}, SecType: {contract.secType}, "
            f"Exchange: {contract.exchange}, Action: {order.action}, OrderType: {order.orderType}, "
            f"TotalQty: {order.totalQuantity}, CashQty: {order.cashQty}, "
            f"LmtPrice: {order.lmtPrice}, "
            f"AuxPrice: {order.auxPrice}, Status: {orderState.status}"
        )
        _logger.info(msg)
        self.broker.log(msg)

        if orderId in self.broker.order_dict.keys():
            order_event = self.broker.order_dict[orderId]
        else:  # not placed by algo
            order_event = InteractiveBrokers.ib_order_to_order(order)
            order_event.order_id = orderId
            order_event.full_symbol = InteractiveBrokers.contract_to_symbol(contract)
            order_event.account = self.broker.account
            order_event.order_time = datetime.now().strftime("%H:%M:%S.%f")
            order_event.source = -1  # unrecognized source
            self.broker.order_dict[orderId] = order_event

        if orderState.status == "Submitted":
            order_event.order_status = OrderStatus.SUBMITTED
        elif orderState.status == "Filled":
            order_event.order_status = OrderStatus.FILLED
        elif orderState.status == "PreSubmitted":
            order_event.order_status = OrderStatus.PENDING_SUBMIT
        elif orderState.status == "Cancelled":
            order_event.order_status = OrderStatus.CANCELED
        elif orderState.status == "Inactive":  # e.g. exchange closed
            order_event.order_status = OrderStatus.ERROR
        else:
            order_event.order_status = OrderStatus.UNKNOWN

        self.broker.event_engine.put(copy(order_event))

    def openOrderEnd(self):
        super().openOrderEnd()
        _logger.info("OpenOrderEnd")
        _logger.info("Received openOrders %s", len(list(self.broker.order_dict.keys())))

    def orderStatus(
        self,
        orderId: OrderId,
        status: str,
        filled: float,
        remaining: float,
        avgFillPrice: float,
        permId: int,
        parentId: int,
        lastFillPrice: float,
        clientId: int,
        whyHeld: str,
        mktCapPrice: float,
    ):
        super().orderStatus(
            orderId,
            status,
            filled,
            remaining,
            avgFillPrice,
            permId,
            parentId,
            lastFillPrice,
            clientId,
            whyHeld,
            mktCapPrice,
        )
        msg = (
            f"OrderStatus. Id: {orderId}, Status: {status}, Filled: {filled}, "
            f"Remaining: {remaining}, AvgFillPrice: {avgFillPrice}, "
            f"PermId: {permId}, ParentId: {parentId}, "
            f"LastFillPrice: {lastFillPrice}, ClientId: {clientId}, "
            f"WhyHeld: {whyHeld}, MktCapPrice: {mktCapPrice}"
        )
        _logger.info(msg)
        self.broker.log(msg)

        order_event = self.broker.order_dict.get(orderId, None)
        if order_event is None:
            msg = f"OrderStatus: Unable to find order {orderId}"
            _logger.error(msg)
            self.broker.log(msg)
            order_event = OrderEvent()
            order_event.order_id = orderId
            order_event.account = self.broker.account
            order_event.order_size = filled + remaining
            order_event.fill_size = filled
            # pylint: disable=attribute-defined-outside-init
            self.order_type = OrderType.UNKNOWN
            self.order_status = OrderStatus.UNKNOWN
            order_event.order_time = datetime.now().strftime("%H:%M:%S.%f")
            order_event.source = -1  # unrecognized source
            self.broker.order_dict[orderId] = order_event

        if status == "Submitted":
            order_event.order_status = OrderStatus.SUBMITTED
        elif status == "Filled":
            order_event.order_status = OrderStatus.FILLED
        elif status == "PreSubmitted":
            order_event.order_status = OrderStatus.PENDING_SUBMIT
        elif status == "Cancelled" or status == "ApiCancelled":
            order_event.order_status = OrderStatus.CANCELED
            order_event.fill_size = filled  # remaining = order_size - fill_size
            order_event.cancel_time = datetime.now().strftime("%H:%M:%S.%f")
        elif status == "Inactive":  # e.g. exchange closed
            order_event.order_status = OrderStatus.ERROR
        else:
            order_event.order_status = OrderStatus.UNKNOWN
        order_event.fill_size = filled

        self.broker.event_engine.put(copy(order_event))

    def managedAccounts(self, accountsList: str):
        super().managedAccounts(accountsList)
        msg = f"Account list:, {accountsList}"
        _logger.info(msg)
        self.broker.log(msg)

        self.broker.account = accountsList.split(",")[0]
        self.reqAccountUpdates(True, self.broker.account)

    def accountSummary(
        self, reqId: int, account: str, tag: str, value: str, currency: str
    ):
        super().accountSummary(reqId, account, tag, value, currency)
        msg = (
            f"AccountSummary. ReqId: {reqId}, Account: {account}, Tag: {tag}, "
            f"Value: {value}, Currency: {currency}"
        )
        _logger.info(msg)
        self.broker.log(msg)

    def accountSummaryEnd(self, reqId: int):
        super().accountSummaryEnd(reqId)
        _logger.info("AccountSummaryEnd. ReqId: %s", reqId)

    def updateAccountValue(self, key: str, val: str, currency: str, accountName: str):
        """
        Just as with the TWS' Account Window, unless there is a position change
        this information is updated at a fixed interval of three minutes.
        """
        super().updateAccountValue(key, val, currency, accountName)
        msg = (
            f"UpdateAccountValue. Key: {key}, Value: {val}, "
            f"Currency: {currency}, AccountName: {accountName}"
        )
        _logger.info(msg)

        self.broker.account_summary.timestamp = datetime.now().strftime("%H:%M:%S.%f")

        if key == "NetLiquidationByCurrency" and currency == "USD":
            self.broker.account_summary.balance = float(val)
        elif key == "NetLiquidation" and currency == "USD":
            self.broker.account_summary.balance = float(val)
            self.broker.account_summary.account_id = accountName
        elif key == "AvailableFunds" and currency == "USD":
            self.broker.account_summary.available = float(val)
        elif key == "MaintMarginReq" and currency == "USD":
            self.broker.account_summary.margin = float(val)
        elif key == "RealizedPnL" and currency == "USD":
            self.broker.account_summary.closed_pnl = float(val)
        elif key == "UnrealizedPnL" and currency == "USD":
            self.broker.account_summary.open_pnl = float(val)
            self.broker.event_engine.put(
                self.broker.account_summary
            )  # assume alphabatic order

    # pylint: disable=too-many-arguments
    def updatePortfolio(
        self,
        contract: Contract,
        position: float,
        marketPrice: float,
        marketValue: float,
        averageCost: float,
        unrealizedPNL: float,
        realizedPNL: float,
        accountName: str,
    ):
        """
        Just as with the TWS' Account Window, unless there is a position change
        this information is updated at a fixed interval of three minutes.
        """
        super().updatePortfolio(
            contract,
            position,
            marketPrice,
            marketValue,
            averageCost,
            unrealizedPNL,
            realizedPNL,
            accountName,
        )
        msg = (
            f"UpdatePortfolio. Symbol: {contract.symbol}, SecType: {contract.secType}, "
            f"Exchange: {contract.exchange}, "
            f"Position: {position}, MarketPrice: {marketPrice}, "
            f"MarketValue: {marketValue}, AverageCost: {averageCost}, "
            f"UnrealizedPNL: {unrealizedPNL}, RealizedPNL: {realizedPNL}, "
            f"AccountName: {accountName}"
        )
        _logger.info(msg)

        position_event = PositionEvent()
        position_event.full_symbol = InteractiveBrokers.contract_to_symbol(contract)
        position_event.sec_type = contract.secType
        position_event.account = accountName
        position_event.size = position
        try:
            multiplier = int(contract.multiplier)
        except ValueError:
            multiplier = 1
        position_event.average_cost = averageCost / multiplier
        position_event.realized_pnl = realizedPNL
        position_event.unrealized_pnl = unrealizedPNL
        position_event.timestamp = datetime.now().strftime("%H:%M:%S.%f")
        self.broker.event_engine.put(position_event)

    def updateAccountTime(self, timeStamp: str):
        super().updateAccountTime(timeStamp)
        msg = f"UpdateAccountTime. Time: {timeStamp}"
        _logger.info(msg)
        self.broker.log(msg)
        # self.broker.event_engine.put(self.broker.account)

    def accountDownloadEnd(self, accountName: str):
        super().accountDownloadEnd(accountName)
        msg = f"AccountDownloadEnd. Account: {accountName}"
        _logger.info(msg)

    def position(
        self, account: str, contract: Contract, position: float, avgCost: float
    ):
        super().position(account, contract, position, avgCost)
        msg = (
            f"Position. Account: {account}, Symbol: {contract.symbol}, "
            f"SecType: {contract.secType}, "
            f"Currency: {contract.currency}, Position: {position}, Avg cost: {avgCost}"
        )
        _logger.info(msg)

    def positionEnd(self):
        super().positionEnd()
        _logger.info("PositionEnd")

    def positionMulti(
        self,
        reqId: int,
        account: str,
        modelCode: str,
        contract: Contract,
        pos: float,
        avgCost: float,
    ):
        super().positionMulti(reqId, account, modelCode, contract, pos, avgCost)
        msg = (
            f"PositionMulti. RequestId: {reqId}, Account: {account}, "
            f"ModelCode: {modelCode}, Symbol: {contract.symbol}, "
            f"SecType: {contract.secType}, Currency: {contract.currency}, Position: {pos}, "
            f"AvgCost: {avgCost}"
        )
        _logger.info(msg)

    def positionMultiEnd(self, reqId: int):
        super().positionMultiEnd(reqId)
        _logger.info("PositionMultiEnd. RequestId: %s", reqId)

    def accountUpdateMulti(
        self,
        reqId: int,
        account: str,
        modelCode: str,
        key: str,
        value: str,
        currency: str,
    ):
        super().accountUpdateMulti(reqId, account, modelCode, key, value, currency)
        msg = (
            f"AccountUpdateMulti. RequestId: {reqId}, Account: {account}, "
            f"ModelCode: {modelCode}, Key: {key}, "
            f"Value: {value}, Currency: {currency}"
        )
        _logger.info(msg)

    def accountUpdateMultiEnd(self, reqId: int):
        super().accountUpdateMultiEnd(reqId)
        _logger.info("AccountUpdateMultiEnd. RequestId: %s", reqId)

    def familyCodes(self, familyCodes: ListOfFamilyCode):
        super().familyCodes(familyCodes)
        _logger.info("Family Codes:")
        for family_code in familyCodes:
            _logger.info("FamilyCode. %s", family_code)

    def pnl(
        self, reqId: int, dailyPnL: float, unrealizedPnL: float, realizedPnL: float
    ):
        super().pnl(reqId, dailyPnL, unrealizedPnL, realizedPnL)
        msg = (
            f"Daily PnL. ReqId: {reqId}, DailyPnL: {dailyPnL}, "
            f"UnrealizedPnL: {unrealizedPnL}, RealizedPnL: {realizedPnL}"
        )
        _logger.info(msg)

    def pnlSingle(
        self,
        reqId: int,
        pos: int,
        dailyPnL: float,
        unrealizedPnL: float,
        realizedPnL: float,
        value: float,
    ):
        super().pnlSingle(reqId, pos, dailyPnL, unrealizedPnL, realizedPnL, value)
        msg = (
            f"Daily PnL Single. ReqId: {reqId}, Position: {pos}, "
            f"DailyPnL: {dailyPnL}, UnrealizedPnL: {unrealizedPnL}, "
            f"RealizedPnL: {realizedPnL}, Value: {value}"
        )
        _logger.info(msg)

    def marketDataType(self, reqId: TickerId, marketDataType: int):
        super().marketDataType(reqId, marketDataType)
        _logger.info("MarketDataType. ReqId: %s, Type: %s", reqId, marketDataType)

    def tickPrice(
        self, reqId: TickerId, tickType: TickType, price: float, attrib: TickAttrib
    ):
        super().tickPrice(reqId, tickType, price, attrib)

        tick_event = self.broker.market_data_tick_dict[reqId]
        tick_event.timestamp = datetime.now()
        if tickType == TickTypeEnum.BID:
            tick_event.tick_type = TickType.BID
            tick_event.bid_price_L1 = price
        elif tickType == TickTypeEnum.ASK:  # pylint: disable=no-member
            tick_event.tick_type = TickType.ASK
            tick_event.ask_price_L1 = price
        elif tickType == TickTypeEnum.LAST:  # pylint: disable=no-member
            tick_event.tick_type = TickType.TRADE
            tick_event.price = price
        else:
            return

        self.broker.tick_event_engine.put(copy(tick_event))

    def tickSize(self, reqId: TickerId, tickType: TickType, size: int):
        super().tickSize(reqId, tickType, size)

        tick_event = self.broker.market_data_tick_dict[reqId]
        tick_event.timestamp = datetime.now()
        if tickType == TickTypeEnum.BID_SIZE:  # pylint: disable=no-member
            tick_event.tick_type = TickType.BID
            tick_event.bid_size_L1 = size
        elif tickType == TickTypeEnum.ASK_SIZE:  # pylint: disable=no-member
            tick_event.tick_type = TickType.ASK
            tick_event.ask_size_L1 = size
        elif tickType == TickTypeEnum.LAST_SIZE:  # pylint: disable=no-member
            tick_event.tick_type = TickType.TRADE
            tick_event.size = size
        else:
            return

        self.broker.tick_event_engine.put(copy(tick_event))

    def tickGeneric(self, reqId: TickerId, tickType: TickType, value: float):
        # pylint: disable=useless-super-delegation
        super().tickGeneric(reqId, tickType, value)
        # _logger.info(f"TickGeneric. TickerId: {reqId}, TickType: {tickType}, Value: {value}")

    def tickString(self, reqId: TickerId, tickType: TickType, value: str):
        # pylint: disable=useless-super-delegation
        super().tickString(reqId, tickType, value)
        # tick_event = self.broker.market_data_tick_dict[reqId]
        # tick_event.timestamp = datetime.fromtimestamp(int(value))
        # self.broker.tick_event_engine.put(copy(tick_event))

    def tickSnapshotEnd(self, reqId: int):
        super().tickSnapshotEnd(reqId)
        msg = f"TickSnapshotEnd. TickerId: {reqId}"
        _logger.info(msg)

    def rerouteMktDataReq(self, reqId: int, conId: int, exchange: str):
        super().rerouteMktDataReq(reqId, conId, exchange)
        msg = f"Re-route market data request. ReqId: {reqId}, ConId: {conId}, Exchange: {exchange}"
        _logger.info(msg)

    def marketRule(self, marketRuleId: int, priceIncrements: ListOfPriceIncrements):
        super().marketRule(marketRuleId, priceIncrements)
        msg = f"Market Rule ID: {marketRuleId}"
        _logger.info(msg)
        for price_increment in priceIncrements:
            msg = f"Price Increment. {price_increment}"
            _logger.info(msg)

    def orderBound(self, reqId: int, apiClientId: int, apiOrderId: int):
        super().orderBound(reqId, apiClientId, apiOrderId)
        msg = f"OrderBound. OrderId: {reqId}, ApiClientId: {apiClientId}, ApiOrderId: {apiOrderId}"
        _logger.info(msg)

    def tickByTickAllLast(
        self,
        reqId: int,
        tickType: int,
        time: int,
        price: float,
        size: int,
        tickAttribLast: TickAttribLast,
        exchange: str,
        specialConditions: str,
    ):
        super().tickByTickAllLast(
            reqId,
            tickType,
            time,
            price,
            size,
            tickAttribLast,
            exchange,
            specialConditions,
        )
        if tickType == 1:
            _logger.info("Last.")
        else:
            _logger.info("AllLast.")
        msg = (
            f'ReqId: {reqId}, Time: {datetime.fromtimestamp(time).strftime("%Y%m%d %H:%M:%S")}, '
            f"Price: {price}, Size: {size}, Exch: {exchange}, Spec Cond: {specialConditions}, "
            f"PastLimit: {tickAttribLast.pastLimit}, Unreported: {tickAttribLast.unreported}"
        )
        _logger.info(msg)

    def tickByTickBidAsk(
        self,
        reqId: int,
        time: int,
        bidPrice: float,
        askPrice: float,
        bidSize: int,
        askSize: int,
        tickAttribBidAsk: TickAttribBidAsk,
    ):
        super().tickByTickBidAsk(
            reqId, time, bidPrice, askPrice, bidSize, askSize, tickAttribBidAsk
        )
        msg = (
            f"BidAsk. ReqId: {reqId}, "
            f'Time: {datetime.fromtimestamp(time).strftime("%Y%m%d %H:%M:%S")}, '
            f"BidPrice: {bidPrice}, AskPrice: {askPrice}, BidSize: {bidSize}, AskSize: {askSize}, "
            f"BidPastLow: {tickAttribBidAsk.bidPastLow}, "
            f"AskPastHigh: {tickAttribBidAsk.askPastHigh}"
        )
        _logger.info(msg)

    def tickByTickMidPoint(self, reqId: int, time: int, midPoint: float):
        super().tickByTickMidPoint(reqId, time, midPoint)
        msg = (
            f"Midpoint. ReqId: {reqId}, "
            f'Time: {datetime.fromtimestamp(time).strftime("%Y%m%d %H:%M:%S")}, '
            f"MidPoint: {midPoint}"
        )
        _logger.info(msg)

    def updateMktDepth(
        self,
        reqId: TickerId,
        position: int,
        operation: int,
        side: int,
        price: float,
        size: int,
    ):
        super().updateMktDepth(reqId, position, operation, side, price, size)
        msg = (
            f"UpdateMarketDepth. ReqId: {reqId}, Position: {position}, "
            f"Operation: {operation}, Side: {side}, Price: {price}, Size: {size}"
        )
        _logger.info(msg)

    def updateMktDepthL2(
        self,
        reqId: TickerId,
        position: int,
        marketMaker: str,
        operation: int,
        side: int,
        price: float,
        size: int,
        isSmartDepth: bool,
    ):
        super().updateMktDepthL2(
            reqId, position, marketMaker, operation, side, price, size, isSmartDepth
        )
        msg = (
            f"UpdateMarketDepthL2. ReqId: {reqId}, "
            f"Position: {position}, MarketMaker: {marketMaker}, "
            f"Operation: {operation}, Side: {side}, Price: {price}, "
            f"Size: {size}, isSmartDepth: {isSmartDepth}"
        )
        _logger.info(msg)

    def rerouteMktDepthReq(self, reqId: int, conId: int, exchange: str):
        super().rerouteMktDataReq(reqId, conId, exchange)
        msg = f"Re-route market depth request. ReqId: {reqId}, ConId: {conId}, Exchange: {exchange}"
        _logger.info(msg)

    def realtimeBar(
        self,
        reqId: TickerId,
        time: int,
        open_: float,
        high: float,
        low: float,
        close: float,
        volume: int,
        wap: float,
        count: int,
    ):
        super().realtimeBar(reqId, time, open_, high, low, close, volume, wap, count)
        msg = f"RealTimeBar. TickerId: {reqId}, time: {time}, close: {close}, count {count}"
        # bar = RealTimeBar(time, -1, open_, high, low, close, volume, wap, count)
        _logger.info(msg)

    def headTimestamp(self, reqId: int, headTimestamp: str):
        msg = f"HeadTimestamp. ReqId: {reqId}, HeadTimeStamp: {headTimestamp}"
        _logger.info(msg)

    def histogramData(self, reqId: int, items: HistogramDataList):
        msg = f"HistogramData. ReqId: {reqId}, HistogramDataList: items[0]"
        _logger.info(msg)

    def historicalData(self, reqId: int, bar: BarData):
        _logger.info("HistoricalData. ReqId: %s, BarData. %s", reqId, bar)

        bar_event = BarEvent()
        bar_event.full_symbol = self.broker.hist_data_request_dict[reqId]
        bar_event.bar_start_time = datetime.strptime(bar.date, "%Y%m%d %H:%M:%S")
        bar_event.interval = 1  # 1 second
        bar_event.open_price = bar.open
        bar_event.high_price = bar.high
        bar_event.low_price = bar.low
        bar_event.close_price = bar.close
        bar_event.volume = bar.volume

        self.broker.tick_event_engine.put(bar_event)

    def historicalDataEnd(self, reqId: int, start: str, end: str):
        super().historicalDataEnd(reqId, start, end)
        msg = f"HistoricalDataEnd. ReqId: {reqId}, from {start}, to {end}"
        _logger.info(msg)

    def historicalDataUpdate(self, reqId: int, bar: BarData):
        msg = f"HistoricalDataUpdate. ReqId: {reqId}, BarData. {bar}"
        _logger.info(msg)

    def historicalTicks(self, reqId: int, ticks: ListOfHistoricalTick, done: bool):
        for tick in ticks:
            _logger.info("HistoricalTick. ReqId: %s, %s", reqId, tick)

    def historicalTicksBidAsk(
        self, reqId: int, ticks: ListOfHistoricalTickBidAsk, done: bool
    ):
        for tick in ticks:
            _logger.info("HistoricalTickBidAsk. ReqId: %s, %s", reqId, tick)

    def historicalTicksLast(
        self, reqId: int, ticks: ListOfHistoricalTickLast, done: bool
    ):
        for tick in ticks:
            _logger.info("HistoricalTickLast. ReqId: %s, %s", reqId, tick)

    def securityDefinitionOptionParameter(
        self,
        reqId: int,
        exchange: str,
        underlyingConId: int,
        tradingClass: str,
        multiplier: str,
        expirations: SetOfString,
        strikes: SetOfFloat,
    ):
        super().securityDefinitionOptionParameter(
            reqId,
            exchange,
            underlyingConId,
            tradingClass,
            multiplier,
            expirations,
            strikes,
        )
        msg = (
            f"SecurityDefinitionOptionParameter. ReqId: {reqId}, "
            f"Exchange: {exchange}, Underlying conId: {underlyingConId}, "
            f"TradingClass: {tradingClass}, Multiplier: {multiplier}, "
            f"Expirations: {expirations}, Strikes: {str(strikes)}"
        )
        _logger.info(msg)

    def securityDefinitionOptionParameterEnd(self, reqId: int):
        super().securityDefinitionOptionParameterEnd(reqId)
        _logger.info("SecurityDefinitionOptionParameterEnd. ReqId: {reqId}")

    def tickOptionComputation(
        self,
        reqId: TickerId,
        tickType: TickType,
        tickAttrib: TickAttrib,
        impliedVol: float,
        delta: float,
        optPrice: float,
        pvDividend: float,
        gamma: float,
        vega: float,
        theta: float,
        undPrice: float,
    ):
        # pylint: disable=arguments-differ
        super().tickOptionComputation(
            reqId,
            tickType,
            tickAttrib,
            impliedVol,
            delta,
            optPrice,
            pvDividend,
            gamma,
            vega,
            theta,
            undPrice,
        )
        msg = (
            f"TickOptionComputation. TickerId: {reqId}, "
            f"TickType: {tickType}, ImpliedVolatility: {impliedVol}, "
            f"Delta: {delta}, OptionPrice: {optPrice}, "
            f"pvDividend: {pvDividend}, Gamma: {gamma}, "
            f"Vega: {vega}, Theta: {theta}, UnderlyingPrice: {undPrice}"
        )
        _logger.info(msg)

    def tickNews(
        self,
        tickerId: int,
        timeStamp: int,
        providerCode: str,
        articleId: str,
        headline: str,
        extraData: str,
    ):
        msg = (
            f"TickNews. TickerId: {tickerId}, "
            f"TimeStamp: {timeStamp}, ProviderCode: {providerCode}, "
            f"ArticleId: {articleId}, Headline: {headline}, ExtraData: {extraData}"
        )
        _logger.info(msg)

    def historicalNews(
        self,
        requestId: int,
        time: str,
        providerCode: str,
        articleId: str,
        headline: str,
    ):
        msg = (
            f"HistoricalNews. ReqId: {requestId}, Time: {time}, ProviderCode: {providerCode}, "
            f"ArticleId: {articleId}, Headline: {headline}"
        )
        _logger.info(msg)

    def historicalNewsEnd(self, requestId: int, hasMore: bool):
        _logger.info("HistoricalNewsEnd. ReqId: %s, HasMore: %s", requestId, hasMore)

    def newsProviders(self, newsProviders: ListOfNewsProviders):
        _logger.info("NewsProviders: ")
        for provider in newsProviders:
            _logger.info("NewsProvider. %s", provider)

    def newsArticle(self, requestId: int, articleType: int, articleText: str):
        msg = (
            f"NewsArticle. ReqId: {requestId}, "
            f"ArticleType: {articleType}, ArticleText: {articleText}"
        )
        _logger.info(msg)

    def contractDetails(self, reqId: int, contractDetails: ContractDetails):
        super().contractDetails(reqId, contractDetails)
        _logger.info(
            "Contract Detail: %s, %s, %s, %s, %s",
            contractDetails.contract.symbol,
            contractDetails.contract.localSymbol,
            contractDetails.contract.primaryExchange,
            contractDetails.contract.exchange,
            contractDetails.contract.multiplier,
        )
        if reqId in self.broker.contract_detail_request_contract_dict.keys():
            self.broker.contract_detail_request_contract_dict[
                reqId
            ] = contractDetails.contract
            self.broker.sym_contract_dict[
                self.broker.contract_detail_request_symbol_dict[reqId]
            ] = contractDetails.contract
            self.broker.contract_symbol_dict[
                contractDetails.contract.conId
            ] = self.broker.contract_detail_request_symbol_dict[reqId]
        else:
            _logger.error(
                "Orphaned contract details request %s, %s",
                reqId,
                contractDetails.underSymbol,
            )

    def bondContractDetails(self, reqId: int, contractDetails: ContractDetails):
        # pylint: disable=useless-super-delegation
        super().bondContractDetails(reqId, contractDetails)

    def contractDetailsEnd(self, reqId: int):
        super().contractDetailsEnd(reqId)
        _logger.info("ContractDetailsEnd. ReqId: %s", reqId)

    def symbolSamples(
        self, reqId: int, contractDescriptions: ListOfContractDescription
    ):
        super().symbolSamples(reqId, contractDescriptions)
        _logger.info("Symbol Samples. Request Id: %s", reqId)

        for contract_description in contractDescriptions:
            deriv_sec_types = ""
            for deriv_sec_type in contract_description.derivativeSecTypes:
                deriv_sec_types += deriv_sec_type
                deriv_sec_types += " "
            _logger.info(
                "Contract: conId:%s, symbol:%s, secType:%s primExchange:%s, "
                "currency:%s, derivativeSecTypes:%s",
                contract_description.contract.conId,
                contract_description.contract.symbol,
                contract_description.contract.secType,
                contract_description.contract.primaryExchange,
                contract_description.contract.currency,
                deriv_sec_types,
            )

    def scannerParameters(self, xml: str):
        super().scannerParameters(xml)
        open("log/scanner.xml", "w", encoding="utf-8").write(xml)
        _logger.info("ScannerParameters received.")

    def scannerData(
        self,
        reqId: int,
        rank: int,
        contractDetails: ContractDetails,
        distance: str,
        benchmark: str,
        projection: str,
        legsStr: str,
    ):
        # pylint: disable=useless-super-delegation
        super().scannerData(
            reqId, rank, contractDetails, distance, benchmark, projection, legsStr
        )

    def scannerDataEnd(self, reqId: int):
        super().scannerDataEnd(reqId)
        _logger.info("ScannerDataEnd. ReqId: %s", reqId)

    def smartComponents(self, reqId: int, smartComponentMap: SmartComponentMap):
        super().smartComponents(reqId, smartComponentMap)
        _logger.info("SmartComponents:")
        for smart_component in smartComponentMap:
            _logger.info("SmartComponent. %s", smart_component)

    def tickReqParams(
        self, tickerId: int, minTick: float, bboExchange: str, snapshotPermissions: int
    ):
        super().tickReqParams(tickerId, minTick, bboExchange, snapshotPermissions)
        msg = (
            f"TickReqParams. TickerId: {tickerId}, MinTick: {minTick}, BboExchange: {bboExchange}, "
            f"SnapshotPermissions: {snapshotPermissions}"
        )
        _logger.info(msg)

    def mktDepthExchanges(self, depthMktDataDescriptions: ListOfDepthExchanges):
        super().mktDepthExchanges(depthMktDataDescriptions)
        _logger.info("MktDepthExchanges:")
        for desc in depthMktDataDescriptions:
            _logger.info("DepthMktDataDescription. %s", desc)

    def fundamentalData(self, reqId: TickerId, data: str):
        super().fundamentalData(reqId, data)
        _logger.info("FundamentalData. ReqId: %s, Data: %s", reqId, data)

    def updateNewsBulletin(
        self, msgId: int, msgType: int, newsMessage: str, originExch: str
    ):
        super().updateNewsBulletin(msgId, msgType, newsMessage, originExch)
        msg = (
            f"News Bulletins. MsgId: {msgId}, Type: {msgType}, "
            f"Message: {newsMessage}, Exchange of Origin: {originExch}"
        )
        _logger.info(msg)

    def receiveFA(self, faData: FaDataType, cxml: str):
        super().receiveFA(faData, cxml)
        _logger.info("Receiving FA: %s", faData)
        open("log/fa.xml", "w", encoding="utf-8").write(cxml)

    def softDollarTiers(self, reqId: int, tiers: list):
        super().softDollarTiers(reqId, tiers)
        _logger.info("SoftDollarTiers. ReqId: %s", reqId)
        for tier in tiers:
            _logger.info("SoftDollarTier. %s", tier)

    def currentTime(self, time: int):
        super().currentTime(time)
        msg = f'CurrentTime: {datetime.fromtimestamp(time).strftime("%H:%M:%S.%f")}'
        _logger.info(msg)
        self.broker.log(msg)

    def execDetails(self, reqId: int, contract: Contract, execution: Execution):
        super().execDetails(reqId, contract, execution)
        msg = (
            f"ExecDetails. ReqId: {reqId}, Symbol: {contract.symbol}, "
            f"SecType: {contract.secType}, "
            f"Currency: {contract.currency}, oid: {execution.orderId}, "
            f"{execution.price}, {execution.shares}"
        )
        _logger.info(msg)

        fill_event = FillEvent()
        fill_event.order_id = execution.orderId
        fill_event.fill_id = execution.execId
        fill_event.fill_price = execution.price
        fill_event.fill_size = execution.shares * (
            1 if execution.side == "BOT" else -1
        )  # BOT SLD
        fill_event.fill_time = datetime.strptime(
            execution.time, "%Y%m%d  %H:%M:%S"
        ).strftime("%H:%M:%S.%f")
        fill_event.exchange = (
            contract.exchange
            if contract.primaryExchange == ""
            else contract.primaryExchange
        )
        fill_event.account = self.broker.account

        if execution.orderId in self.broker.order_dict.keys():
            fill_event.full_symbol = self.broker.order_dict[
                execution.orderId
            ].full_symbol
            fill_event.source = self.broker.order_dict[execution.orderId].source
        else:  # not placed by algo
            fill_event.full_symbol = InteractiveBrokers.contract_to_symbol(contract)

        self.broker.event_engine.put(fill_event)

    def execDetailsEnd(self, reqId: int):
        super().execDetailsEnd(reqId)
        _logger(f"ExecDetailsEnd. ReqId: {reqId}")

    def displayGroupList(self, reqId: int, groups: str):
        super().displayGroupList(reqId, groups)
        _logger.info("DisplayGroupList. ReqId: %s, Groups: %s", reqId, groups)

    def displayGroupUpdated(self, reqId: int, contractInfo: str):
        super().displayGroupUpdated(reqId, contractInfo)
        _logger.info(
            "DisplayGroupUpdated. ReqId: %s, ContractInfo: %s", reqId, contractInfo
        )

    def commissionReport(self, commissionReport: CommissionReport):
        super().commissionReport(commissionReport)
        _logger.info("CommissionReport. %s", commissionReport)

    def completedOrder(self, contract: Contract, order: Order, orderState: OrderState):
        super().completedOrder(contract, order, orderState)
        msg = (
            f"CompletedOrder. PermId: {order.permId}, "
            f"ParentPermId: {utils.longToStr(order.parentPermId)}, "
            f"Account: {order.account}, Symbol: {contract.symbol}, "
            f"SecType: {contract.secType}, Exchange: {contract.exchange}, "
            f"Action: {order.action}, OrderType: {order.orderType}, "
            f"TotalQty: {order.totalQuantity}, "
            f"CashQty: {order.cashQty}, FilledQty: {order.filledQuantity}, "
            f"LmtPrice: {order.lmtPrice}, "
            f"AuxPrice: {order.auxPrice}, Status: {orderState.status}, "
            f"Completed time: {orderState.completedTime}, "
            f"Completed Status: {orderState.completedStatus}"
        )
        _logger.info(msg)

    def completedOrdersEnd(self):
        super().completedOrdersEnd()
        _logger.info("CompletedOrdersEnd")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import numpy as np
import pandas as pd
from datetime import datetime, date
import logging

from .event import EventType
from .event.backtest_event_engine import BacktestEventEngine
from .data.backtest_data_feed import BacktestDataFeed
from .data.data_board import DataBoard
from .brokerage.backtest_brokerage import BacktestBrokerage
from .position.position_manager import PositionManager
from .order.order_manager import OrderManager
from .performance.performance_manager import PerformanceManager
from .risk.risk_manager import PassThroughRiskManager
from .strategy import StrategyManager

_logger = logging.getLogger(__name__)


class BacktestEngine(object):
    """
    Event driven backtest engine
    """

    def __init__(self, start_date=None, end_date=None):
        self._current_time = None
        self._start_date = start_date
        self._end_date = end_date
        self.config = dict()
        self.config[
            "strategy"
        ] = dict()  # to be consistent with live; in backtest, strategy is set outside
        self.instrument_meta = {}  # one copy of meta dict shared across program
        self._data_feed = BacktestDataFeed(self._start_date, self._end_date)
        self._data_board = DataBoard()
        self._performance_manager = PerformanceManager(
            self.instrument_meta
        )  # send dict pointer
        self._position_manager = PositionManager("Global")
        self._position_manager.set_instrument_meta(self.instrument_meta)
        self._order_manager = OrderManager("Global")
        self._events_engine = BacktestEventEngine(self._data_feed)
        self._backtest_brokerage = BacktestBrokerage(
            self._events_engine, self._data_board
        )
        self._risk_manager = PassThroughRiskManager()
        self._strategy_manager = StrategyManager(
            self.config,
            self._backtest_brokerage,
            self._order_manager,
            self._position_manager,
            self._risk_manager,
            self._data_board,
            self.instrument_meta,
        )
        self._strategy = None

    def set_instrument_meta(self, instrument_meta):
        self.instrument_meta.update(instrument_meta)

    def set_capital(self, capital):
        """
        set capital to the global position manager
        """
        self._position_manager.set_capital(capital)

    def set_strategy(self, strategy):
        self._strategy = strategy

    def add_data(self, data_key, data_source, watch=True):
        """
        Add data for backtest
        :param data_key: AAPL or CL
        :param data_source:  dataframe, datetimeindex
        :param watch: track position or not
        :return:
        """
        if data_key not in self.instrument_meta.keys():
            keys = data_key.split(" ")
            # find first digit position
            for i, c in enumerate(keys[0]):
                if c.isdigit():
                    break
                if i < len(keys[0]):
                    sym_root = keys[0][: i - 1]
                    if sym_root in self.instrument_meta.keys():
                        self.instrument_meta[data_key] = self.instrument_meta[sym_root]

        self._data_feed.set_data_source(data_source)  # get iter(datetimeindex)
        self._data_board.initialize_hist_data(data_key, data_source)
        if watch:
            self._performance_manager.add_watch(data_key, data_source)

    def _setup(self):
        """
        Tis needs to be run after strategy and data are loaded
        because it subscribes to market data
        """
        ## 1. data_feed
        self._data_feed.subscribe_market_data()

        ## 4. set strategy
        self._strategy.active = True
        self._strategy_manager.load_strategy({self._strategy.name: self._strategy})

        ## 5. global performance manager and portfolio manager
        self._performance_manager.reset()
        self._position_manager.reset()

        ## 5. trade recorder
        # self._trade_recorder = ExampleTradeRecorder(output_dir)

        ## 6. wire up event handlers
        self._events_engine.register_handler(EventType.TICK, self._tick_event_handler)
        # to be consistent with current live, order is placed directly; this accepts other status like status, fill, cancel
        self._events_engine.register_handler(EventType.ORDER, self._order_event_handler)
        self._events_engine.register_handler(EventType.FILL, self._fill_event_handler)

    # ------------------------------------ private functions -----------------------------#
    def _tick_event_handler(self, tick_event):
        self._current_time = tick_event.timestamp

        # performance update goes before position and databoard updates because it updates previous day performance
        # it can't update today because orders haven't been filled yet.
        self._performance_manager.update_performance(
            self._current_time, self._position_manager, self._data_board
        )
        self._position_manager.mark_to_market(
            tick_event.timestamp,
            tick_event.full_symbol,
            tick_event.price,
            self._data_board,
        )
        self._strategy.on_tick(
            tick_event
        )  # plus strategy.position_manager market to marekt
        # data_baord update after strategy, so it still holds price of last tick; for position MtM
        # strategy uses tick.price for current price; and use data_board.last_price for previous price
        # for backtest, this is PLACEHOLDER based on timestamp.
        # strategy pull directly from data_board hist_data for current_price; and data_board.last_price for previous price
        self._data_board.on_tick(tick_event)
        # check standing orders, after databoard is updated
        self._backtest_brokerage.on_tick(tick_event)

    def _order_event_handler(self, order_event):
        """
        acknowledge order
        """
        # self._backtest_brokerage.place_order(order_event)
        self._order_manager.on_order_status(order_event)
        self._strategy.on_order_status(order_event)
        pass

    def _fill_event_handler(self, fill_event):
        self._order_manager.on_fill(fill_event)
        self._position_manager.on_fill(fill_event)
        self._performance_manager.on_fill(fill_event)
        self._strategy.on_fill(fill_event)

    # -------------------------------- end of private functions -----------------------------#
    def run(self):
        """
        Run backtest
        """
        self._setup()

        self._events_engine.run()
        # explicitly update last day/time
        self._performance_manager.update_performance(
            self._current_time, self._position_manager, self._data_board
        )

        return (
            self._performance_manager._equity,
            self._performance_manager._df_positions,
            self._performance_manager._df_trades,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .order_type import *
from .order_status import *
from datetime import datetime
from copy import copy
import logging

_logger = logging.getLogger(__name__)


class OrderManager(object):
    """
    Manage/track all the orders
    """

    def __init__(self, name="Global"):
        self.name = name
        self.order_dict = {}  # order_id ==> order
        self.fill_dict = {}  # fill_id ==> fill
        self.standing_order_set = set()  # order_id of standing order for convenience
        self.canceled_order_set = set()  # order_id of canceled orders for convenience

    def reset(self):
        self.order_dict.clear()
        self.fill_dict.clear()
        self.standing_order_set.clear()
        self.canceled_order_set.clear()

    def on_tick(self, tick_event):
        """ """
        pass

    def on_order_status(self, order_event):
        """
        on order status change from broker
        including canceled status
        """
        # there should be no negative order id if order is directly placed without queue.
        if order_event.order_id < 0:
            _logger.error(
                f"{self.name} OrderManager received negative orderid {order_event.order_id}"
            )

        if order_event.order_id in self.order_dict:
            if (
                order_event.full_symbol
                != self.order_dict[order_event.order_id].full_symbol
            ):
                _logger.error(f"{self.name} OrderManager Error: orders dont match")
                return False
            # only change status when it is logical
            elif (
                self.order_dict[order_event.order_id].order_status.value
                <= order_event.order_status.value
            ):
                self.order_dict[
                    order_event.order_id
                ].order_status = order_event.order_status
                if order_event.order_status < OrderStatus.FILLED:
                    self.standing_order_set.add(order_event.order_id)
                elif order_event.order_status == OrderStatus.CANCELED:
                    self.canceled_order_set.add(order_event.order_id)
                    self.order_dict[
                        order_event.order_id
                    ].cancel_time = order_event.cancel_time
                    if order_event.order_id in self.standing_order_set:
                        self.standing_order_set.remove(order_event.order_id)
                return True
            else:  # no need to change status
                return False
        # order_id not yet assigned, open order at connection or placed by trader?
        else:
            self.order_dict[order_event.order_id] = copy(
                order_event
            )  # it is important to use copy
            if order_event.order_status < OrderStatus.FILLED:
                self.standing_order_set.add(order_event.order_id)
            elif order_event.order_status == OrderStatus.CANCELED:
                self.canceled_order_set.add(order_event.order_id)
                if order_event.order_id in self.standing_order_set:
                    self.standing_order_set.remove(order_event.order_id)
            return True

    def on_cancel(self, oid):
        """
        This proactively set order status to PENDING_CANCEL
        :param o:
        :return:
        """
        # Cancel will be handled in order_status
        # self.canceled_order_set.add(oid)
        # if oid in self.standing_order_set:
        #     self.standing_order_set.remove(oid)
        if oid in self.order_dict.keys():
            self.order_dict[oid].order_status = OrderStatus.PENDING_CANCEL
        else:
            _logger.error(f"{self.name} OrderManager cancel order is not registered")

    def on_fill(self, fill_event):
        """
        on receive fill_event from broker
        """
        if fill_event.fill_id in self.fill_dict.keys():
            _logger.error(f"{self.name} fill exists")
        else:
            self.fill_dict[fill_event.fill_id] = fill_event

            if fill_event.order_id in self.order_dict:
                self.order_dict[fill_event.order_id].fill_price = (
                    fill_event.fill_price * fill_event.fill_size
                    + self.order_dict[fill_event.order_id].fill_price
                    * self.order_dict[fill_event.order_id].fill_size
                ) / (
                    self.order_dict[fill_event.order_id].fill_size
                    + fill_event.fill_size
                )
                self.order_dict[fill_event.order_id].fill_size += fill_event.fill_size

                if (
                    self.order_dict[fill_event.order_id].order_size
                    == self.order_dict[fill_event.order_id].fill_size
                ):
                    self.order_dict[
                        fill_event.order_id
                    ].order_status = OrderStatus.FILLED
                    if fill_event.order_id in self.standing_order_set:
                        self.standing_order_set.remove(fill_event.order_id)
                else:
                    self.order_dict[
                        fill_event.order_id
                    ].order_status = OrderStatus.PARTIALLY_FILLED
            else:
                _logger.error(
                    f"{self.name} Fill event {fill_event.fill_id} has no matching order {fill_event.order_id}"
                )

    def retrieve_order(self, order_id):
        try:
            return self.order_dict[order_id]
        except:
            return None

    def retrieve_fill(self, fill_id):
        try:
            return self.fill_dict[fill_id]
        except:
            return None

    def retrieve_standing_orders(self):
        oids = []
        for oid in self.standing_order_set:
            if oid in self.order_dict.keys():
                if (
                    self.order_dict[oid].order_status < OrderStatus.FILLED
                ):  # has standing order
                    oids.append(oid)
        return oids



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

_logger = logging.getLogger(__name__)


class Position(object):
    def __init__(self, full_symbol, average_price, size, realized_pnl=0):
        """
        Position includes zero/closed security
        """
        ## TODO: add cumulative_commission, long_trades, short_trades, round_trip etc
        self.full_symbol = full_symbol
        # average price includes commission
        self.average_price = average_price
        self.size = size
        self.realized_pnl = 0
        self.unrealized_pnl = 0
        self.account = ""

    def get_current_pnl(self):
        return self.realized_pnl, self.unrealized_pnl

    def mark_to_market(self, last_price, multiplier):
        """
        given new market price, update the position
        """
        # if long or size > 0, pnl is positive if last_price > average_price
        # else if short or size < 0, pnl is positive if last_price < average_price
        self.unrealized_pnl = (last_price - self.average_price) * self.size * multiplier

    def on_fill(self, fill_event, multiplier, parent_name):
        """
        adjust average_price and size according to new fill/trade/transaction
        """
        if self.full_symbol != fill_event.full_symbol:
            _logger.error(
                "%s Position symbol %s and fill event symbol %s do not match. "
                % (parent_name, self.full_symbol, fill_event.full_symbol)
            )

        if self.size > 0:  # existing long
            if fill_event.fill_size > 0:  # long more
                self.average_price = (
                    self.average_price * self.size
                    + fill_event.fill_price * fill_event.fill_size
                    + fill_event.commission / multiplier
                ) / (self.size + fill_event.fill_size)
            else:  # flat long
                # _logger.info(f'{parent_name} flat long realized_pnl {self.realized_pnl}, avg {self.average_price}, fill {fill_event.fill_price}, {fill_event.fill_size}, {multiplier}, {fill_event.commission}')
                if abs(self.size) >= abs(fill_event.fill_size):  # stay long
                    self.realized_pnl += (
                        self.average_price - fill_event.fill_price
                    ) * fill_event.fill_size * multiplier - fill_event.commission
                else:  # flip to short
                    self.realized_pnl += (
                        fill_event.fill_price - self.average_price
                    ) * self.size * multiplier - fill_event.commission
                    self.average_price = fill_event.fill_price
        elif self.size < 0:  # existing short
            if fill_event.fill_size < 0:  # short more
                self.average_price = (
                    self.average_price * self.size
                    + fill_event.fill_price * fill_event.fill_size
                    + fill_event.commission / multiplier
                ) / (self.size + fill_event.fill_size)
            else:  # flat short
                # _logger.info(f'{parent_name} flat short realized_pnl {self.realized_pnl}, avg {self.average_price}, fill {fill_event.fill_price}, {fill_event.fill_size}, {multiplier}, {fill_event.commission}')
                if abs(self.size) >= abs(fill_event.fill_size):  # stay short
                    self.realized_pnl += (
                        self.average_price - fill_event.fill_price
                    ) * fill_event.fill_size * multiplier - fill_event.commission
                else:  # flip to long
                    self.realized_pnl += (
                        fill_event.fill_price - self.average_price
                    ) * self.size * multiplier - fill_event.commission
                    self.average_price = fill_event.fill_price
        else:  # no position
            self.average_price = (
                fill_event.fill_price
                + fill_event.commission / multiplier / fill_event.fill_size
            )

        self.size += fill_event.fill_size

        _logger.info(
            f"{parent_name} Position Fill: sym {self.full_symbol}, avg price {self.average_price}, fill price {fill_event.fill_price}, fill size {fill_event.fill_size}, after size {self.size}, close pnl {self.realized_pnl}"
        )



"""
Definition of the constants of the module
"""
from methodtools import lru_cache

from .client import Client, LAST_MODIFIED as last_modified

EMPTY = "empty"

FUTURE_TYPE = "Future"


@lru_cache()
def get_futures():
    """
    Get the futures.
    """
    futures, _ = Client().get_tickers()
    return futures


LAST_MODIFIED = last_modified

LETTERS = ["F", "G", "H", "J", "K", "M", "N", "Q", "U", "V", "X", "Z"]

LIBOR_BEFORE_2001 = 6.65125



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


class RiskManagerBase(metaclass=ABCMeta):
    """
    RiskManager base class
    """

    @abstractmethod
    def order_in_compliance(self, o, strategy_manager=None):
        raise NotImplementedError("order_in_compliance should be implemented")



"""
Data feed base module.
"""
from abc import ABCMeta, abstractmethod


class DataFeedBase(metaclass=ABCMeta):
    """
    DateFeed baae class
    """

    @abstractmethod
    def subscribe_market_data(self, symbols):
        """subscribe to market data"""

    @abstractmethod
    def unsubscribe_market_data(self, symbols):
        """unsubscribe market data"""

    @abstractmethod
    def stream_next(self):
        """stream next data event"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from datetime import datetime, timedelta
from ..order.order_status import OrderStatus
from ..order.order_event import OrderEvent

_logger = logging.getLogger(__name__)


class StrategyManager(object):
    def __init__(
        self,
        config,
        broker,
        order_manager,
        position_manager,
        risk_manager,
        data_board,
        instrument_meta,
    ):
        """
        current design: oversees all strategies/traders, check with risk managers before send out orders
        let strategy manager to track strategy position for each strategy, with the help from order manager

        :param config:
        :param strat_dict:     strat name ==> stract
        :param broker: to place order directly without message queue
        :param order_manager:  this order manager support total/global orders
        :param position_manager:    this position manager help track total/global positions
        :param risk_manager: chck order witth
        :param data_board: for strategy
        """
        self._config = config
        self._broker = broker
        self._order_manager = order_manager
        self._position_manager = position_manager
        self._risk_manager = risk_manager
        self._data_board = data_board
        self._strategy_dict = {}  # sid ==> strategy
        self._instrument_meta = instrument_meta  # symbol ==> instrument_meta
        self._tick_strategy_dict = {}  # sym -> list of strategy
        self._sid_oid_dict = {
            0: [],
            -1: [],
        }  # sid ==> oid list; 0: manual; -1: unknown source

    def load_strategy(self, strat_dict):
        sid = 1  # 0 is mannual discretionary trade, or not found
        # similar to backtest; strategy sets capital, params, and symbols
        for k, v in strat_dict.items():
            v.id = sid
            sid += 1
            v.name = k
            if v.name in self._config["strategy"].keys():
                v.active = self._config["strategy"][v.name]["active"]
                v.set_capital(self._config["strategy"][v.name]["capital"])  # float
                v.set_params(self._config["strategy"][v.name]["params"])  # dict
                v.set_symbols(self._config["strategy"][v.name]["symbols"])  # list

                # yaml converts to seconds
                if "order_start_time" in self._config["strategy"][v.name].keys():
                    if isinstance(
                        self._config["strategy"][v.name]["order_start_time"], int
                    ):
                        self._config["strategy"][v.name]["order_start_time"] = str(
                            timedelta(
                                seconds=self._config["strategy"][v.name][
                                    "order_start_time"
                                ]
                            )
                        )
                if "order_end_time" in self._config["strategy"][v.name].keys():
                    if isinstance(
                        self._config["strategy"][v.name]["order_end_time"], int
                    ):
                        self._config["strategy"][v.name]["order_end_time"] = str(
                            timedelta(
                                seconds=self._config["strategy"][v.name][
                                    "order_end_time"
                                ]
                            )
                        )

            self._strategy_dict[v.id] = v
            self._sid_oid_dict[v.id] = []  # record its orders
            for sym in v.symbols:
                if sym not in self._instrument_meta.keys():
                    # find first digit position
                    ss = sym.split(" ")
                    for i, c in enumerate(ss[0]):
                        if c.isdigit():
                            break
                    if i < len(ss[0]):
                        sym_root = ss[0][: i - 1]
                        if sym_root in self._instrument_meta.keys():
                            self._instrument_meta[sym] = self._instrument_meta[
                                sym_root
                            ]  # add for quick access

                if sym in self._tick_strategy_dict:
                    self._tick_strategy_dict[sym].append(v.id)
                else:
                    self._tick_strategy_dict[sym] = [v.id]
                if sym in self._broker.market_data_subscription_reverse_dict.keys():
                    continue
                else:
                    _logger.info(f"add {sym}")
                    self._broker.market_data_subscription_reverse_dict[sym] = -1

            v.on_init(self, self._data_board, self._instrument_meta)

    def start_strategy(self, sid):
        self._strategy_dict[sid].active = True

    def stop_strategy(self, sid):
        self._strategy_dict[sid].active = False

    def pause_strategy(self, sid):
        self._strategy_dict[sid].active = False

    def start_all(self):
        for k, v in self._strategy_dict.items():
            v.active = True

    def stop_all(self):
        for k, v in self._strategy_dict.items():
            v.active = False

    def place_order(self, o, check_risk=True):
        # currently it puts order directly with broker; e.g. by simplying calling ib.placeOrder method
        # Because order is placed directly; all subsequent on_order messages are order status updates
        # TODO, use an outbound queue to send orders
        # 1. check with risk manager
        order_check = True
        if check_risk:
            order_check = self._risk_manager.order_in_compliance(o, self)

        # 2. if green light
        if not order_check:
            return

        # 2.a record
        oid = self._broker.orderid
        self._broker.orderid += 1
        o.order_id = oid
        o.order_status = OrderStatus.NEWBORN
        self._sid_oid_dict[o.source].append(oid)
        # feedback newborn status
        self._order_manager.on_order_status(o)
        if (
            o.source in self._strategy_dict.keys()
        ):  # in case it is not placed by strategy
            self._strategy_dict[o.source].on_order_status(o)

        # 2.b place order
        self._broker.place_order(o)

    def cancel_order(self, oid):
        self._order_manager.on_cancel(oid)
        # self._strategy_dict[sid].on_cancel(oid)  # This is moved to strategy_base
        self._broker.cancel_order(oid)

    def cancel_strategy(self, sid):
        """
        call strategy cancel to take care of strategy order_manager
        """
        if sid not in self._strategy_dict.keys():
            _logger.error(f"Cancel strategy can not locate strategy id {sid}")
        else:
            self._strategy_dict[sid].cancel_all()

    def cancel_all(self):
        for sid, s in self._strategy_dict.items():
            s.cancel_all()

    def flat_strategy(self, sid):
        """
        flat with MARKET order (default)
        Assume each strategy track its own positions
        TODO: should turn off strategy?
        """
        if sid not in self._strategy_dict.keys():
            _logger.error(f"Flat strategy can not locate strategy id {sid}")

        for sym, pos in self._strategy_dict[sid]._position_manager.positions.items():
            if pos.size != 0:
                o = OrderEvent()
                o.full_symbol = sym
                o.order_size = -pos.size
                o.source = 0  # mannual flat
                o.create_time = datetime.now().strftime("%H:%M:%S.%f")
                self.place_order(
                    o, check_risk=False
                )  # flat strategy doesnot cehck risk

    def flat_all(self):
        """
        flat all according to position_manager
        TODO: should turn off all strategies?
        :return:
        """
        for sym, pos in self._position_manager.positions.items():
            if pos.size != 0:
                o = OrderEvent()
                o.full_symbol = sym
                o.order_size = -pos.size
                o.source = 0
                o.create_time = datetime.now().strftime("%H:%M:%S.%f")
                self.place_order(
                    o, check_risk=False
                )  # flat strategy doesnot cehck risk

    def on_tick(self, k):
        if k.full_symbol in self._tick_strategy_dict.keys():
            s_list = self._tick_strategy_dict[k.full_symbol]
            for sid in s_list:
                if self._strategy_dict[sid].active:
                    self._strategy_dict[sid].on_tick(k)

    def on_position(self, pos):
        """
        get initial position
        read from config file instead
        :param pos:
        :return:
        """
        pass

    def on_order_status(self, order_event):
        """
        TODO: check if source is working
        :param order_event:
        :return:
        """
        sid = order_event.source
        if sid in self._strategy_dict.keys():
            self._strategy_dict[sid].on_order_status(order_event)
        else:
            _logger.info(
                f"strategy manager doesnt hold the oid {order_event.order_id} to set status {order_event.order_status}, possibly from outside of the system"
            )

    def on_cancel(self, order_event):
        """
        TODO no need for this
        """
        sid = order_event.source
        if sid in self._strategy_dict.keys():
            self._strategy_dict[sid].on_order_status(order_event)
        else:
            _logger.info(
                f"strategy manager doesnt hold the oid {order_event.order_id} to cancel, possibly from outside of the system"
            )

    def on_fill(self, fill_event):
        """
        assign fill ordering to order id ==> strategy id
        TODO: check fill_event source; if not, fix it or use fill_event.order_id
        """
        sid = fill_event.source
        if sid in self._strategy_dict.keys():
            self._strategy_dict[sid].on_fill(fill_event)
        else:
            _logger.info(
                f"strategy manager doesnt hold the oid {fill_event.order_id} to fill, possibly from outside of the system"
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod


class AbstractTradeRecorder(object):
    """
    transaction recorder
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def record_trade(self, fill):
        """
        logs fill event
        """
        raise NotImplementedError("Should implement record_trade()")



"""
Tick event module.
"""
from datetime import datetime
from enum import Enum

from pandas import Timestamp

from ..event.event import Event, EventType


class TickType(Enum):
    """
    Unlike IB, it does not have tick_size, e.g., TickTypeEnum.BID_SIZE
    """

    TRADE = 0
    BID = 1
    ASK = 2
    FULL = 3


class TickEvent(Event):
    """
    Tick event
    """

    def __init__(self):
        """
        Initialises Tick
        """
        self.event_type = EventType.TICK
        self.tick_type = TickType.TRADE
        self.timestamp = Timestamp("1970-01-01", tz="UTC")
        self.full_symbol = ""
        self.price = 0.0
        self.size = 0
        self.depth = 1
        self.bid_price_L1 = 0.0  # pylint: disable=invalid-name
        self.bid_size_L1 = 0  # pylint: disable=invalid-name
        self.ask_price_L1 = 0.0  # pylint: disable=invalid-name
        self.ask_size_L1 = 0  # pylint: disable=invalid-name
        self.open_interest = 0
        self.open = 0.0
        self.high = 0.0
        self.low = 0.0
        self.pre_close = 0.0
        self.upper_limit_price = 0.0
        self.lower_limit_price = 0.0

    def __str__(self):
        return (
            f'{self.timestamp.strftime("%H:%M:%S.%f")},'
            f'{datetime.now().strftime("%H:%M:%S.%f")},{self.full_symbol},'
            f"{self.tick_type},{self.bid_size_L1},{self.bid_price_L1},"
            f"{self.ask_price_L1},{self.ask_size_L1},{self.price},{self.size}"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
from ..position import Position
import logging

_logger = logging.getLogger(__name__)


class PositionManager(object):
    def __init__(self, name):
        """ """
        self.name = name
        self.initial_capital = 0
        self.cash = 0
        # current total value after market to market, before trades from strategy.
        # After-trades calculated in performanace manager
        self.current_total_capital = 0
        self.contracts = {}  # symbol ==> contract
        self.positions = {}  # symbol ==> positions
        self.instrument_meta = {}  # sym ==> instrument_meta

    def set_instrument_meta(self, instrument_meta_dict):
        self.instrument_meta = instrument_meta_dict

    def set_capital(self, initial_capital):
        self.initial_capital = initial_capital

    def reset(self):
        self.cash = self.initial_capital
        self.current_total_capital = self.initial_capital
        self.contracts.clear()
        self.positions.clear()

    def get_holdings_count(self):
        n = 0
        for s, p in self.positions.items():
            if p.size != 0:
                n += 1
        return n

    def get_position_size(self, symbol):
        if symbol in self.positions.keys():
            return self.positions[symbol].size
        else:
            return 0

    def get_cash(self):
        return self.cash

    def get_total_pnl(self):
        total_pnl = 0
        for s, pos in self.positions.items():
            cp, op = pos.get_current_pnl()
            total_pnl = total_pnl + cp + op
        return total_pnl

    def on_contract(self, contract):
        if contract.full_symbol not in self.contracts:
            self.contracts[contract.full_symbol] = contract
            _logger.info(
                f"{self.name} Contract {contract.full_symbol} information received. "
            )
        else:
            _logger.info(
                f"{self.name} Contract {contract.full_symbol} information already exists "
            )

    def on_position(self, pos_event):
        """respond to updatePortfolio; global position_manager only"""
        pos = pos_event.to_position()
        self.positions[pos.full_symbol] = pos

    def on_fill(self, fill_event):
        """
        This works only on stocks.
        TODO: consider margin
        """
        # sell will get cash back
        if fill_event.full_symbol in self.instrument_meta.keys():
            multiplier = self.instrument_meta[fill_event.full_symbol]["Multiplier"]
        else:
            multiplier = 1

        self.cash -= (
            fill_event.fill_size * fill_event.fill_price
        ) * multiplier + fill_event.commission
        self.current_total_capital -= fill_event.commission  # commission is a cost

        if fill_event.full_symbol in self.positions:  # adjust existing position
            self.positions[fill_event.full_symbol].on_fill(
                fill_event, multiplier, self.name
            )
        else:
            self.positions[fill_event.full_symbol] = fill_event.to_position(multiplier)

    def mark_to_market(self, time_stamp, symbol, last_price, data_board):
        """
        from previous timestamp to current timestamp. Pnl from holdings
        """
        if symbol == "PLACEHOLDER":  # backtest placeholder, update all
            for sym, pos in self.positions.items():
                if sym in self.instrument_meta.keys():
                    multiplier = self.instrument_meta[sym]["Multiplier"]
                else:
                    multiplier = 1
                real_last_price = data_board.get_current_price(
                    sym, time_stamp
                )  # not PLACEHOLDER
                pos.mark_to_market(real_last_price, multiplier)
                # data board not updated yet; get_last_time return previous time_stamp
                self.current_total_capital += (
                    self.positions[sym].size
                    * (real_last_price - data_board.get_last_price(sym))
                    * multiplier
                )
        elif symbol in self.positions:
            # this is a quick way based on one symbol; actual pnl should sum up across positions
            if symbol in self.instrument_meta.keys():
                multiplier = self.instrument_meta[symbol]["Multiplier"]
            else:
                multiplier = 1

            self.positions[symbol].mark_to_market(last_price, multiplier)
            prev_price = data_board.get_last_price(symbol)
            if prev_price is not None:  # in case data board hasn't been updated/empty
                self.current_total_capital += (
                    self.positions[symbol].size * (last_price - prev_price) * multiplier
                )
